### BRAZILLIAN EXCHANGE MARKET DATA

import ast
import os
import sys
import investpy
import math
import json
import yfinance as yf
import pandas as pd
import numpy as np
from yahooquery import Ticker
from requests import request
from datetime import date, datetime
from typing import Tuple, Optional, List, Union
sys.path.append('\\'.join([d for d in os.path.dirname(
    os.path.realpath(__file__)).split('\\')][:-3]))
from stpstone.settings._global_slots import YAML_B3
from stpstone.cals.handling_dates import DatesBR
from stpstone.handling_data.json import JsonFiles
from stpstone.handling_data.str import StrHandler
from stpstone.finance.comdinheiro.api_request import ComDinheiro
from stpstone.handling_data.object import HandlingObjects
from stpstone.cals.handling_dates import DatesBR
from stpstone.handling_data.lists import HandlingLists
from stpstone.document_numbers.br import DocumentsNumbersBR
from stpstone.handling_data.dicts import HandlingDicts
from stpstone.handling_data.html import HtmlHndler
from stpstone.handling_data.numbers import NumHandler
from stpstone.loggs.db_logs import DBLogs


class TradingVolumeB3:

    def bov_monthly(
            self, 
            int_year:int, 
            int_month:int
        ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # setting variables
        list_th = list()
        list_td = list()
        # format month and year to string
        str_month = f'{int_month:02}'
        str_year = str(int_year)
        url = YAML_B3['bov_trading_volume']['url'].format(str_month, str_year)
        # request html
        bs_html = HtmlHndler().html_bs_parser(
            url, 
            bl_verify=YAML_B3['bov_trading_volume']['bl_verify']
        )
        # table
        bs_table = bs_html.find_all('table')[11]
        # looping within rows
        for i_tr, tr in enumerate(bs_table.find_all('tr')):
            #   checking if is a header how, otherwise it is a data row
            if i_tr < 2:
                #   getting headers
                list_th.extend([
                    StrHandler().remove_diacritics(el.get_text())
                        .replace('\xa0', '')
                        .replace('Totais dos pregoes  Ref: ', '')
                        .replace('(R$)', 'BRL')
                        .replace(' - ', ' ')
                        .replace(' ', '_')
                        .strip()
                        .upper()
                    for el in tr.find_all('td')
                    if len(
                        StrHandler().remove_diacritics(el.get_text())\
                        .replace('\xa0', '')
                    ) > 0
                ])
            else:
                #   getting data
                list_td.extend([
                    # data numeric
                    float(td.get_text()
                        .strip()
                        .replace('.', '')
                        .replace(',', '.'))
                    if NumHandler().is_numeric(
                        StrHandler().remove_diacritics(td.get_text())
                            .strip()
                            .replace('.', '')
                            .replace(',', '.')
                    )
                    # data not numeric
                    else
                        StrHandler().remove_diacritics(td.get_text())
                            .strip()
                            .replace(' de ', ' ')
                            .replace(' do ', ' ')
                            .replace(' a ', ' ')
                            .replace(' e ', ' ')
                            .replace(' - ', ' ')
                            .replace('-', ' ')
                            .replace(' / ', ' ')
                            .replace(' ', '_')
                            .replace('.', '')
                            .replace(',', '.')
                            .replace('(', '')
                            .replace(')', '')
                            .replace('/', '')
                            .upper()
                    for td in tr.find_all('td')
                ])
        # dealing with header raw data
        list_th = [
            list_th[2],
            list_th[3] + '_' + list_th[0],
            list_th[4] + '_' + list_th[0],
            list_th[3] + '_' + list_th[1],
            list_th[4] + '_' + list_th[1],
        ]
        # pair headers and data within a list
        list_ser = HandlingDicts().pair_headers_with_data(
            list_th, 
            list_td
        )
        # turning into dataframe
        df_bov_montlhy = pd.DataFrame(list_ser)
        # defining columns of month and last twelve months
        cols_month = list(df_bov_montlhy.columns)[0:3]
        col_l12m = [list(df_bov_montlhy.columns)[0]]
        col_l12m.extend(
            list(df_bov_montlhy.columns)[3:]
        )
        # creating separate into two dataframes
        df_bov_mnt = df_bov_montlhy[cols_month]
        df_bov_l12m = df_bov_montlhy[cols_month]
        # renaming columns
        for df_ in [
            df_bov_mnt,
            df_bov_l12m
        ]:
            #   list columns
            list_cols = list(df_.columns)
            #   renaming columns of interest
            df_.rename(columns={
                list_cols[1]: YAML_B3['bov_trading_volume']['col_deals'],
                list_cols[2]: YAML_B3['bov_trading_volume']['col_volume'],
                list_cols[0]: YAML_B3['bov_trading_volume']['col_market']
            }, inplace=True)
            #   adding column of time period
            df_[YAML_B3['bov_trading_volume']['col_time_period']] = \
                DatesBR().month_year_string(
                    list_cols[1].replace('NEGOCIACOES_', '')
                )
            # adding logging
            df_ = DBLogs().audit_log(
                df_, url, list_cols[1].replace('NEGOCIACOES_', '')
            )
        # returning dataframe
        return df_bov_mnt, df_bov_l12m


class CalendarB3:

    def bl_weekly_option(self, ticker_opc:str, int_min_letters:int=6) -> bool:
        '''
        DOCSTRING: BOOLEAN TO WEEKLY EXPIRING OPTIONS
        INPUTS: TICKER
        OUTPUTS: BOOLEAN
        '''
        num_let = sum(c.isalpha() for c in ticker_opc)
        num_let += sum(c.isdigit() for c in ticker_opc[:4])
        return num_let >= int_min_letters

    @property
    def options_exercise_dates(self) -> pd.DataFrame:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # setting variables
        list_ser = list()
        # request html
        bs_html = HtmlHndler().html_bs_parser(
            YAML_B3['options_exercise_dates']['url'], 
            bl_verify=YAML_B3['options_exercise_dates']['bl_verify']
        )
        # looping within li tag
        for bs_li in bs_html.find_all('li'):
            #   getting month
            try:
                str_month = bs_li.find('a').get_text()
            except AttributeError:
                continue
            #   getting headers
            list_headers = [x.get_text().upper() for x in bs_li.find_all('th')]
            #   getting data
            list_data = [x.get_text() for x in bs_li.find_all('td')]
            #   checking headers and data list, in case of empty data continue
            if len(list_headers) == 0 or len(list_data) == 0: continue
            #   pair headers and data within a list
            list_ser = HandlingDicts().pair_headers_with_data(
                list_headers, 
                list_data
            )
            #   adding month to serialized list
            list_ser = HandlingDicts().add_k_v_serialized_list(
                list_ser, 
                YAML_B3['options_exercise_dates']['key_month'].upper(), 
                str_month
            )
            #   exporting to serialized list
            list_ser.extend(list_ser)
        # exporting serialized list to dataframe
        df_opt_xcr_dts = pd.DataFrame(list_ser)
        # adding request date
        df_opt_xcr_dts[YAML_B3['options_exercise_dates']['key_request_date'].upper()] = \
            DatesBR().curr_date
        # altering data types
        df_opt_xcr_dts = df_opt_xcr_dts.astype({
            YAML_B3['options_exercise_dates']['key_day'].upper(): int,
            YAML_B3['options_exercise_dates']['key_details'].upper(): str,
            YAML_B3['options_exercise_dates']['key_month'].upper(): str
        })
        # adding logging
        df_opt_xcr_dts = DBLogs().audit_log(
            df_opt_xcr_dts, YAML_B3['options_exercise_dates']['url'], 
            DatesBR().utc_log_ts
        )
        # returning dataframe
        return df_opt_xcr_dts


class TradingHoursB3:

    def futures_generic(self, url:str, int_cols:int) -> list:
        '''
        DOCSTRING: TRADING TIMES OF FUTURES REGARDING BRAZILLIAN PMI (IPCA) AND STOCK INDEXES
        INPUTS: -
        OUTPUTS:
        '''
        # setting variables
        list_df = list()
        # request html
        bs_html = HtmlHndler().html_bs_parser(
            url, 
            bl_verify=YAML_B3['trading_hours_b3']['futures']['bl_verify']
        )
        # looping within tables
        for _, bs_table in enumerate(bs_html.find_all('table')):
            #   reseting variables
            list_ser = list()
            #   creating headers' list
            #   auxiliary header 1
            list_headers_1 = [
                StrHandler().remove_diacritics(bs_th.get_text())\
                    .upper()\
                    .replace('(', '')\
                    .replace(')', '')\
                    .replace(' ', '_')\
                    .replace('+', '')\
                    .replace('-', '_')\
                    .replace('_DE_', '_')
                for bs_th in bs_table.find('thead').find_all('tr')[0].find_all('th')
            ]
            #   checking wheter there is more than one column for the header
            try:
                #   auxiliary header 2
                list_headers_2 = [
                    StrHandler().remove_diacritics(bs_td.get_text())\
                        .upper()
                    for bs_td in bs_table.find('thead').find_all('tr')[1].find_all('td')
                ]
                #   consolidated header
                list_headers = [
                    list_headers_1[0], 
                    list_headers_1[1]
                ]
                for aux_ in range(2, int_cols):
                    if aux_ < 4:
                        list_headers.extend(
                            [
                                list_headers_2[2*aux_-4] + '_' + list_headers_1[aux_],
                                list_headers_2[2*aux_-3] + '_' + list_headers_1[aux_]
                            ]
                        )
                    elif aux_ == 4:
                        list_headers.extend(
                            [
                                list_headers_2[2*aux_-4] + '_' + list_headers_1[aux_]
                            ]
                        )
                    else:
                        list_headers.extend(
                            [
                                list_headers_2[2*aux_-5] + '_' + list_headers_1[aux_],
                                list_headers_2[2*aux_-4] + '_' + list_headers_1[aux_]
                            ]
                        )
            except IndexError:
                list_headers = list_headers_1
            #   looping within rows
            for bs_row in bs_table.find('tbody').find_all('tr'):
                #   body
                list_tr_body = [
                    bs_td.get_text()\
                        .replace('(1)', '')\
                        .replace('(2)', '')\
                        .replace('(3)', '')\
                        .replace('(4)', '')\
                        .replace('(5)', '')\
                        .replace('(6)', '')\
                        .replace('(7)', '')\
                        .replace('\n', '')
                    for bs_td in bs_row.find_all('td')
                ]
                #   appending to serialized list
                list_ser.append(dict(zip(list_headers, list_tr_body)))
            #   creating dataframe
            df_ = pd.DataFrame(list_ser)
            #   appending to list
            list_df.append(df_)
            #   adding logging
            df_ = DBLogs().audit_log(
                df_, url, 
                DatesBR().utc_log_ts
            )
        # returning dataframes
        return list_df

    @property
    def futures_pmi_idx(self) -> list:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return self.futures_generic(
            YAML_B3['trading_hours_b3']['futures']['url_pmi_idx'], 
            YAML_B3['trading_hours_b3']['futures']['num_cols_pmi_idx']
        )

    @property
    def futures_brl_usd_int_rts(self) -> list:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return self.futures_generic(
            YAML_B3['trading_hours_b3']['futures']['url_int_rts'], 
            YAML_B3['trading_hours_b3']['futures']['num_cols_int_rts']
        )
    
    @property
    def futures_commodities(self) -> list:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS
        '''
        return self.futures_generic(
            YAML_B3['trading_hours_b3']['futures']['url_commodities'], 
            YAML_B3['trading_hours_b3']['futures']['num_cols_commodities']
        )
    
    @property
    def futures_crypto(self) -> list:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return self.futures_generic(
            YAML_B3['trading_hours_b3']['futures']['url_crypto'], 
            YAML_B3['trading_hours_b3']['futures']['num_cols_crypto']
        )
    
    @property
    def futures_forex(self) -> list:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS
        '''
        return self.futures_generic(
            YAML_B3['trading_hours_b3']['futures']['url_forex'], 
            YAML_B3['trading_hours_b3']['futures']['num_cols_forex']
        )

    @property
    def futures_otc(self) -> list:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return self.futures_generic(
            YAML_B3['trading_hours_b3']['futures']['url_otc'], 
            YAML_B3['trading_hours_b3']['futures']['num_cols_otc']
        )
    
    @property
    def futures_opf_bef_aft_xrc(self) -> list:
        '''
        DOCSTRING: OPTIONS ON FUTURES - BEFORE AND AFTER EXERCISE
        INPUTS: -
        OUTPUTS: 
        '''
        return self.futures_generic(
            YAML_B3['trading_hours_b3']['futures']['url_opf'], 
            YAML_B3['trading_hours_b3']['futures']['num_cols_opf']
        )

    @property
    def stocks(self) -> list:
        '''
        DOCSTRING: TRADING TIMES OF BRAZILLIAN STOCKS
        INPUTS: -
        OUTPUTS:
        '''
        # setting variables
        list_df = list()
        # request html
        bs_html = HtmlHndler().html_bs_parser(
            YAML_B3['trading_hours_b3']['stocks']['url'], 
            bl_verify=YAML_B3['trading_hours_b3']['stocks']['bl_verify']
        )
        # looping within tables
        for i, bs_table in enumerate(bs_html.find_all('table')):
            #   reseting variables
            list_ser = list()
            #   CREATING HEADRERS' LIST
            #   TABLE 0
            if i == 0:
                #   auxiliary header 1
                list_headers_1 = [
                    StrHandler().remove_numeric_chars(
                        StrHandler().remove_diacritics(bs_th.get_text())\
                            .upper()\
                            .replace(' ', '_')\
                            .replace('-', '_')\
                            .replace('_DE_', '_')
                    )
                    for bs_th in bs_table.find_all('th')
                    if StrHandler().find_substr_str(
                        StrHandler().remove_diacritics(bs_th.get_text())\
                            .upper()\
                            .replace(' ', '_')\
                            .replace('-', '_')\
                            .replace('_DE_', ''),
                        '*AFTER_MARKET*'
                    ) == False
                ]
                list_headers_1.remove('AFTER_MARKET')
                list_headers_1[-1] = list_headers_1[-1] + '_AFTER_MARKET'
                list_headers_1[-2] = list_headers_1[-2] + '_AFTER_MARKET'
                #   auxiliary header 2
                list_headers_2 = [
                    StrHandler().remove_numeric_chars(
                        StrHandler().remove_diacritics(bs_td.get_text())\
                            .upper()
                    )
                    for bs_td in bs_table.find('tbody').find('tr').find_all('td')
                ]
                #   consolidating headers list
                list_headers = [
                    list_headers_1[0]
                ]
                for aux_ in range(1, 8):
                    list_headers.extend(
                        [
                            list_headers_2[2*aux_-1] + '_' + list_headers_1[aux_],
                            list_headers_2[2*aux_] + '_' + list_headers_1[aux_]
                        ]
                    )
            #   TABLE 1
            if i == 1:
                #   auxiliary header 1
                list_headers_1 = [
                    StrHandler().remove_numeric_chars(
                        StrHandler().remove_diacritics(bs_th.get_text())\
                            .upper()\
                            .replace(' ', '_')\
                            .replace('-', '_')\
                            .replace('_DO_', '_')
                    )
                    for bs_th in bs_table.find('thead').find_all('tr')[1].find_all('th')
                ]
                #   auxiliary header 2
                list_headers_2 = [
                    StrHandler().remove_numeric_chars(
                        StrHandler().remove_diacritics(bs_th.get_text())\
                            .upper()\
                            .replace(' ', '_')\
                            .replace('-', '_')\
                            .replace('_DO_', '_')\
                            .replace('_DE_', '_')
                    )
                    for bs_th in bs_table.find('thead').find_all('tr')[2].find_all('th')
                ]
                #   auxiliary header 3
                list_headers_3 = [
                    StrHandler().remove_diacritics(bs_td.get_text())\
                        .upper()\
                        .replace('\xa0', '')
                    for bs_td in bs_table.find('tbody').find_all('tr')[0].find_all('td')
                ]
                #   consolidating headers list
                list_headers = [
                    'MERCADO', 
                    list_headers_3[1] + '_' + list_headers_2[0] \
                                + '_' + list_headers_1[0],
                    list_headers_3[2] + '_' + list_headers_2[0] \
                        + '_' + list_headers_1[0], 
                    list_headers_3[3] + '_' + list_headers_2[1] \
                                + '_' + list_headers_1[1],
                    list_headers_3[4] + '_' + list_headers_2[1] \
                        + '_' + list_headers_1[1], 
                    list_headers_2[2] + '_' + list_headers_1[1], 
                    list_headers_3[6] + '_' + list_headers_2[3] \
                        + '_' + list_headers_1[1], 
                    list_headers_3[7] + '_' + list_headers_2[3] \
                        + '_' + list_headers_1[1], 
                    list_headers_3[8] + '_' + list_headers_2[4] \
                        + '_' + list_headers_1[1], 
                ]
            else:
                raise Exception('Table {} not expected, please validate.'.format(i))
            #   looping within rows
            for bs_row in bs_table.find('tbody').find_all('tr')[1:]:
                #   body
                list_tr_body = [
                    bs_td.get_text()\
                        .replace(r'\xa', '')\
                        .replace('\xa0','')\
                        .replace('\n', '')\
                        .replace('\t', '')\
                        .replace('(1)', '')\
                        .replace('(2)', '')
                    for bs_td in bs_row.find_all('td')
                ]
                #   appending to serialized list
                list_ser.append(dict(zip(list_headers, list_tr_body)))
            #   creating dataframe
            df_ = pd.DataFrame(list_ser)
            #   appending to list
            list_df.append(df_)
            #   adding logging
            df_ = DBLogs().audit_log(
                df_, YAML_B3['trading_hours_b3']['stocks']['url'], 
                DatesBR().utc_log_ts
            )
        # creating individual dataframes
        df_tt_bov = list_df[0]
        df_tt_exc_opt = list_df[1]
        # returning dataframes of interest
        return df_tt_bov, df_tt_exc_opt


class TheorPortfB3:

    def generic_req(self, str_indice:str, method:str='GET', float_pct_factor:float=100.0) \
        -> pd.DataFrame:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # requesting data
        resp_req = request(method, YAML_B3['theor_port_b3']['url_{}'.format(str_indice)], 
                           verify=YAML_B3['theor_port_b3']['bl_verify'])
        # raising exception in case of status code different from 2xx
        resp_req.raise_for_status()
        # requesting json
        json_ibov = resp_req.json()
        # building serialized list
        list_ser = [
            {
                YAML_B3['theor_port_b3']['col_ticker']: dict_[YAML_B3[
                    'theor_port_b3']['key_code']], 
                YAML_B3['theor_port_b3']['col_asset']: dict_[YAML_B3[
                    'theor_port_b3']['key_asset']], 
                YAML_B3['theor_port_b3']['col_type']: dict_[YAML_B3[
                    'theor_port_b3']['key_type']], 
                YAML_B3['theor_port_b3']['col_pct']: float(dict_[YAML_B3[
                    'theor_port_b3']['key_part']].replace(',', '.')) / float_pct_factor, 
                YAML_B3['theor_port_b3']['col_theor_qty']: dict_[YAML_B3[
                    'theor_port_b3']['key_theor_qty']].replace('.', ''), 
            } for dict_ in json_ibov['results']
        ]
        # defining dataframe
        df_ = pd.DataFrame(list_ser)
        # changing data types
        df_ = df_.astype({
            YAML_B3['theor_port_b3']['col_ticker']: str, 
            YAML_B3['theor_port_b3']['col_asset']: str, 
            YAML_B3['theor_port_b3']['col_type']: str, 
            YAML_B3['theor_port_b3']['col_pct']: float, 
            YAML_B3['theor_port_b3']['col_theor_qty']: float
        })
        # adding logging
        df_ = DBLogs().audit_log(
            df_, YAML_B3['theor_port_b3']['url_{}'.format(str_indice)], 
            DatesBR().utc_log_ts
        )
        # returning dataframe
        return df_
    
    @property
    def ibov(self) -> pd.DataFrame:
        ''''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return self.generic_req('ibov')
    
    @property
    def ibra(self) -> pd.DataFrame:
        ''''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return self.generic_req('ibra')
    
    @property
    def ibrx100(self) -> pd.DataFrame:
        ''''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return self.generic_req('ibrx100')
    
    @property
    def ibrx50(self) -> pd.DataFrame:
        ''''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return self.generic_req('ibrx50')


class MDB3:

    @property
    def financial_indicators_b3(self, method:str='GET', float_pct_factor:float=100.0, 
                                dt_input_fmt:str='DD/MM/YYYY') -> pd.DataFrame:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # requesting financial indicators b3
        resp_req = request(method, YAML_B3['financial_indicators']['url'])
        # raise exception if status code is different from 2xx
        resp_req.raise_for_status()
        # appeding json to dataframe
        df_fin_ind = pd.DataFrame(resp_req.json())
        # changing column types
        df_fin_ind = df_fin_ind.astype({
            YAML_B3['financial_indicators']['col_si_code']: np.int64,
            YAML_B3['financial_indicators']['col_desc']: str,
            YAML_B3['financial_indicators']['col_grp_desc']: str,
            YAML_B3['financial_indicators']['col_value']: str,
            YAML_B3['financial_indicators']['col_rate']: str,
            YAML_B3['financial_indicators']['col_last_up']: str,
        })
        df_fin_ind[YAML_B3['financial_indicators']['col_last_up']] = [
            DatesBR().str_date_to_datetime(x, dt_input_fmt) if x is not math.nan else 0 
            for x in df_fin_ind[YAML_B3['financial_indicators']['col_last_up']]
        ]
        for col_ in [
            YAML_B3['financial_indicators']['col_value'],
            YAML_B3['financial_indicators']['col_rate']
        ]:
            df_fin_ind[col_] = [
                float(x.replace(',', '.')) if x != '' else 0 
                for x in df_fin_ind[col_]
            ]
        df_fin_ind[YAML_B3['financial_indicators']['col_rate']] = [
            #   selic rate
            (row[YAML_B3['financial_indicators']['col_rate']] + YAML_B3['financial_indicators'][
                'float_pct_dif_cdi_selic']) / float_pct_factor if row[YAML_B3[
                    'financial_indicators']['col_si_code']] == YAML_B3['financial_indicators'][
                        'int_si_code_selic']
            #   other rates
            else row[YAML_B3['financial_indicators']['col_rate']] / float_pct_factor
            for _, row in df_fin_ind.iterrows()
        ]
        # adding logging
        df_fin_ind = DBLogs().audit_log(
            df_fin_ind, 
            YAML_B3['financial_indicators']['url'],
            DatesBR().utc_from_dt(DatesBR().curr_date)
        )
        # returning dataframe
        return df_fin_ind

    @property
    def securities_volatility(self, method:str='GET', float_pct_factor:float=100.0) -> pd.DataFrame:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # setting variables
        list_ser = list()
        # looping through page sets
        for _, dict_ in YAML_B3['securities_volatility_b3']['pages'].items():
            #   looping through hashes
            for str_hash in dict_['hashes']:
                #   request security volatility page
                resp_req = request(method, dict_['url'].format(str_hash))
                #   raise exception for status different from 2xx
                resp_req.raise_for_status()
                #   get json in memory
                json_ = resp_req.json()
                #   reseting json with results
                list_ = json_['results']
                #   creating serialized list
                for k, v in [
                    (YAML_B3['securities_volatility_b3']['cols_names']['pg_num'], 
                     json_['page']['pageNumber']), 
                    (YAML_B3['securities_volatility_b3']['cols_names']['total_pgs'], 
                     json_['page']['totalPages']), 
                    (YAML_B3['securities_volatility_b3']['cols_names']['date_ref'], 
                     DatesBR().curr_date), 
                     (YAML_B3['securities_volatility_b3']['cols_names']['url'], 
                     dict_['url']), 
                ]:
                    list_ = HandlingDicts().add_k_v_serialized_list(list_, k , v)
                #   extend to serialized list
                list_ser.extend(list_)
        # adding to pandas dataframe
        df_sec_vol = pd.DataFrame(list_ser)
        # changing column types
        df_sec_vol = df_sec_vol.astype({
            YAML_B3['securities_volatility_b3']['cols_names']['ticker']: str,
            YAML_B3['securities_volatility_b3']['cols_names']['company_name']: str,
            YAML_B3['securities_volatility_b3']['cols_names']['serie']: str, 
            YAML_B3['securities_volatility_b3']['cols_names']['pg_num']: int,
            YAML_B3['securities_volatility_b3']['cols_names']['total_pgs']: int
        })
        #   looping within periods of security volatility calculed by b3
        for int_per in YAML_B3['securities_volatility_b3']['vols_calc_per']:
            #   loop through standard deviation and annual volatility for the giver period
            for col_ in [
                YAML_B3['securities_volatility_b3']['cols_names']['std'].format(str(int_per)), 
                YAML_B3['securities_volatility_b3']['cols_names']['ann_vol'].format(str(int_per))
            ]:
                #   alter dataframe column types
                df_sec_vol[col_] = [float(str(d).replace(',', '.')) / float_pct_factor 
                                    for d in df_sec_vol[col_]]
        # adding logging
        df_sec_vol = DBLogs().audit_log(
            df_sec_vol, None, DatesBR().utc_log_ts
        )
        # returning dataframe
        return df_sec_vol


class MDYFinance:

    def historical_data_securities_yq(self, dt_inf:datetime, dt_sup:datetime, 
                                      list_securities:Optional[list]=None, 
                                      list_indexes:Optional[list]=None,
                                      bl_verify:bool=False, 
                                      column_ticker:str='ticker', 
                                      colum_dt_date:str='dt_date') -> pd.DataFrame:
        '''
        DOCSTRING:
        INPTUS:
        OUTPUTS:
        '''
        # dealing with securities list_tickers
        if (list_securities == None) or (list_securities == []):
            list_securities = []
        else:
            # list_securities = ['{}.SA'.format(str(x).upper()) for x in list_securities]
            list_securities = [str(x).upper() for x in list_securities]
        # dealing with indexes list_tickers
        if (list_indexes == None) or (list_indexes == []):
            list_indexes = []
        else:
            list_indexes = ['^{}'.format(str(x).upper()) for x in list_indexes]
        # extending stock's list
        list_tickers = HandlingLists().extend_lists(list_indexes, list_securities)
        # removing duplicates
        list_tickers = HandlingLists().remove_duplicates(list_tickers)
        # getting historical data
        list_yq_data = Ticker(list_tickers, verify=bl_verify)
        df_yq_data = list_yq_data.history(start=dt_inf, end=dt_sup)
        # index to column - ticker
        df_yq_data[column_ticker] = [str(x[0]).replace('^', '').replace('.SA', '') 
                                     for x in df_yq_data.index]
        # index to column - date in datetime format
        df_yq_data[colum_dt_date] = [x[1] for x in df_yq_data.index]
        # filling nan values with upper data
        df_yq_data[column_ticker].fillna(method='ffill', inplace=True)
        # retrieving historical data
        return df_yq_data

    def daily_returns(self, df_yq_data:pd.DataFrame, column_symbol:str='symbol', 
                      col_ticker:str='ticker', col_close:str='adjclose', 
                      col_open:str='open', col_date:str='dt_date', 
                      col_daily_return:str='daily_return', 
                      str_type_return_calc:str='close_close') -> pd.DataFrame:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # order by ticker/date in ascending order
        df_yq_data.sort_values(
            by=[col_ticker, col_date], 
            ascending=[True, True], 
            inplace=True
        )
        # daily returns grouped by symbol
        if str_type_return_calc == 'close_close':
            df_yq_data[col_daily_return] = (
                df_yq_data.groupby(col_ticker)[col_close]
                .apply(lambda x: np.log(x / x.shift(1)))
                .reset_index(0, drop=True)
            )
        elif str_type_return_calc == 'open_close':
            df_yq_data[col_daily_return] = (
                df_yq_data.groupby(col_ticker).apply(
                    lambda group: np.log(group[col_open] / group[col_close].shift(1))
                ).reset_index(level=0, drop=True)
            )
        else:
            raise Exception('Type of return calculation not supported. '
                            + f'TYPE: {str_type_return_calc}')
        # optionally reset nulls when switching symbols for clarity
        df_yq_data.loc[
            df_yq_data.groupby(level=column_symbol).head(1).index, col_daily_return
        ] = None
        # returning dataframe with daily returns
        return df_yq_data


class MDInvestingDotCom:

    def ticker_reference_investing_com(
        self, str_ticker:str, 
        str_host:str='https://tvc4.investing.com/725910b675af9252224ca6069a1e73cc/1631836267/1/1/8/symbols?symbol={}', 
        str_method:str='GET', bl_verify=True, key_ticker='ticker', 
        dict_headers:dict={
            'User-Agent': 'Mozilla/5.0 (Windowns NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.101 Safari/537.36'
        }) -> json:
        '''
        DOCSTRING: TICKER REFERENCE FROM INVESTING.COM
        INPUTS: TICKER, HOST (DEFAULT), METHOD (DEFAULT), BOOLEAN VERIFY (DEFAULT), 
            KEY TICKER (DEFAULT), HEADERS (DEFAULT)
        OUTPUTS: STRING
        '''
        # collect content from rest
        req_resp = request(
            str_method, 
            str_host.format(str_ticker), 
            verify=bl_verify, 
            headers=dict_headers
        )
        req_resp.raise_for_status()
        # turning to desired format - loading json to memopry
        json_ = req_resp.json()
        # return named ticker form investing.com
        return json_[key_ticker]

    def historical_closing_intraday_data_investing_com(
        self, str_ticker_ref:str, dt_inf:datetime,
        dt_sup:datetime, str_type_closing_intraday:str='D',
        str_host:str='https://tvc4.investing.com/725910b675af9252224ca6069a1e73cc/'
            + '1631836267/1/1/8/history?symbol={}&resolution={}&from={}&to={}', str_method:str='GET',
        dict_headers:dict={
            'User-Agent': 'Mozilla/5.0 (Windowns NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.101 Safari/537.36'
        }) -> json:
        '''
        DOCSTRING: HISTORICAL CLOSING/INTRADAY TICKS FROM INVESTING.COM
        INPUTS: TICKER REFERENCE FROM INVESTING.COM, FROM DATE (TIMESTAMP), TO DATE (TIMESTAMP), 
            TYPE CLOSING/INTRADAY PRICE (D AS DEFAULT FOR CLOSING), HOST (DEFAULT) AND HEADERS
        OUTPUTS: JSON
        '''
        # collect content from rest
        req_resp = request(
            str_method, 
            str_host.format(
                str_ticker_ref, 
                str_type_closing_intraday,
                dt_inf.strftime('%Y-%m-%d'), 
                dt_sup.strftime('%Y-%m-%d')
            ),
            headers=dict_headers
        )
        req_resp.raise_for_status()
        # turning to desired format - loading json to memory
        json_ = req_resp.json()
        # return named ticker from investing.com
        return json_

    def historical_closing_data_yf(
            self, 
            list_tickers:List[str], 
            str_period:str='max',
            col_date:str='Date', 
            col_ticker:str='Ticker'
        ) -> pd.DataFrame:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS
        '''
        # setting variables
        list_ser = list()
        # looping through list_tickers and collecting historical data
        for ticker in list_tickers:
            #   creating yfinance object for the current ticker
            obj_yf = yf.Ticker(ticker)
            #   collecting historical data
            df_hist_closing_data = obj_yf.history(period=str_period)
            #   creating date column, which is the index
            df_hist_closing_data[col_date] = df_hist_closing_data.index
            #   creating column with ticker name
            df_hist_closing_data[col_ticker] = ticker
            #   appending to exporting list of dictionaries, which will be converted into a dataframe
            list_ser.extend(df_hist_closing_data.to_dict(orient='records'))
        # alterando lista de dicionários para dataframe
        df_closing_data = pd.DataFrame(list_ser)
        # alterando tipo de coluna de data para string
        df_closing_data[col_date] = [StrHandler().get_string_until_substring(str(x), ' ') 
                                     for x in df_closing_data[col_date]]
        # alterando string para tipo de data
        df_closing_data[col_date] = [DatesBR().str_date_to_datetime(x, 'YYYY-MM-DD') 
                                     for x in df_closing_data[col_date]]
        # returning data of interest
        return df_closing_data

    def close_price_hist(
            self, list_tickers:List[str], dt_inf:datetime, dt_sup:datetime, 
            str_country:str='brazil', str_assets_class:str='acao'
        ) -> pd.DataFrame:
        '''
        DOCSTRING: HISTORIC CLOSING PRICES FROM INVESTING.COM
        INPUTS: TICKERS, DATA INFERIOR, DATA SUPERIOR, PAÍS (BRAZIL COMO DEFAULT) COM
            FORMATO DATA 'DD/MM/YYYY'
        OUTPUTS: JSON
        '''
        # setting variables
        dict_close = dict()
        # filling list of available assets
        if str_assets_class == 'acao' and type(str_assets_class) == str:
            for ticker in list_tickers:
                json_ = HandlingObjects().literal_eval_data(
                    investpy.get_stock_historical_data(ticker, str_country, dt_inf, dt_sup,
                                                       as_json=True))
                dict_close[ticker] = json_
        else:
            if type(str_assets_class) == str:
                str_assets_class = [str_assets_class]
                str_assets_class.extend((len(list_tickers) - 1) * [str_assets_class])
            dict_ticker_name = dict()
            for i in range(len(list_tickers)):
                if str_assets_class[i] == 'etf':
                    for dict_ativo in investpy.get_etfs_dict(str_country):
                        if dict_ativo['symbol'] == list_tickers[i]:
                            dict_ticker_name[list_tickers[i]
                                                    ] = dict_ativo['full_name']
                            break
                    else:
                        dict_ticker_name[list_tickers[i]] = \
                            'Ticker não cadastrado no serviço de cotações investing.com'
                    json_ = investpy.get_etf_historical_data(
                        dict_ticker_name[list_tickers[i]],
                        str_country, dt_inf, dt_sup, as_json=True)
                    dict_close[list_tickers[i]] = HandlingObjects().literal_eval_data(
                        json_)
                elif str_assets_class[i] == 'acao':
                    json_ = investpy.get_stock_historical_data(
                        list_tickers[i], str_country, dt_inf, dt_sup, as_json=True)
                    dict_close[list_tickers[i]] = HandlingObjects().literal_eval_data(
                        json_)
        return pd.DataFrame(dict_close)

    def idx_hist_data(
            self, 
            dt_inf:datetime, 
            dt_sup:datetime, 
            str_idx:str='Bovespa', 
            str_country:str='brazil'
        ) -> pd.DataFrame:
        '''
        DOCSTRING: CLOSING PRICE OF A GIVEN INDICE
        INPUTS: DATE INFERIOR, SUPERIOR, INDICE NAME (DEFAULT BOVESPA), 
            COUNTRY (DEFAULT BRAZIL), FORMAT OF EXTRACTION (DEFAULT JSON2)
        OUTPUTS: DICT WITH CLOSE PRICE
        '''
        return pd.DataFrame([
            {d['date']: d['close']} 
            for d in HandlingObjects().literal_eval_data(
                investpy.indices.get_index_historical_data(
                    str_idx, str_country, dt_inf, dt_sup, as_json=True
               ))['historical']
        ])


class MDComDinheiro:

    def __init__(self, user:str, passw:str) -> None:
        self.user = user
        self.passw = passw

    def bmf_historical_close_data(
            self, 
            str_ticker:str, 
            str_maturity_code:str,
            dt_inf:datetime, 
            dt_sup:datetime, 
            str_fmt_extract:str='json2'
        ) -> json:
        '''
        DOCSTRING: CLOSING PRICE OF BMF CONTRACTS
        INPUTS: USERNAME (COMDINHEIRO), PASSWORD (COMDINHEIRO), CONTRACT CODE, MATURITY CODE, 
            DATE INFERIOR (DDMMAAAA, AS A STRING, OR DATETIME FORMAT), DATE SUPERIOR (SAME FORMAT 
            AS DATE INFERIOR) AND FORMAT EXTRACTION (JSON AS DEFAULT)
        OUTPUTS: JSON
        '''
        # applying date format
        if DatesBR().check_date_datetime_format(dt_inf) == True:
            dt_inf = DatesBR().datetime_to_string(dt_inf, '%d%m%Y')
        if DatesBR().check_date_datetime_format(dt_sup) == True:
            dt_sup = DatesBR().datetime_to_string(dt_sup, '%d%m%Y')
        # str_payload - self.user / password / str_ticker name / inferior date / superior date / maturity
        #   code / format
        str_payload = str(
            f'username={self.user}&password={self.passw}&URL=HistoricoCotacaoBMF'
            + f'-{str_ticker}-{dt_inf}-{dt_sup}-{str_maturity_code}-1&format={str_fmt_extract}'
        )
        # sending rest resp_req
        json_ = ComDinheiro().requests_api_cd(str_payload).text.encode('utf8')
        json_ = HandlingObjects().literal_eval_data(json_, "b'", "'").replace(r'\n', '')
        return json_

    def indice_neg(self, list_tickers:List[str], dt_inf:datetime, dt_sup:datetime) -> json:
        '''
        DOCSTRING: FUNÇÃO PARA TRAZER O ÍNDICE DE NEGOCIABILIDADE DA B3
        INPUTS: ATIVO, DATA INICIAL E DATA FINAL ('DD/MM/AAAA')
        OUTPUTS: JSON COM O VALOR DO ÍNDICE DE NEGOCIABILIDADE NO PERÍODO
        '''
        # definindo variáveis para pool de conexão
        if len(list_tickers) > 1:
            str_list_tickers = '%2B'.join(list_tickers)
        else:
            str_list_tickers = ''.join(list_tickers)
        dt_inf = dt_inf.strftime('%d~%m~%Y')
        dt_sup = dt_sup.strftime('%d~%m~%Y')
        str_payload = 'username={}&password={}&URL=ComparaEmpresas001.php%3F%26d'.format(
            self.user, self.passw) \
            + 'ata_d%3D31129999%26data_a%3D16%2F06%2F2020%26trailing%3D12%26conv%3DMIXED' \
            + '%26c_c%3Dconsolidado%2520preferencialmente%26moeda%3DMOEDA_ORIGINAL' \
            + '%26m_m%3D1000000000%26n_c%3D2%26f_v%3D1%26' \
            + 'papeis%3D' \
            + str_list_tickers \
            + '%26indic%3DNEGOCIABILIDADE(' \
            + dt_inf \
            + '%2C' \
            + dt_sup \
            + '%2C%2C%2C2)' \
            + '%26enviar_email%3D0%26enviar_email_log%3D0%26transpor%3D0%26op01%3D' \
            + 'tabela%26oculta_cabecalho_sup%3D0%26relat_alias_automatico%3' \
            + 'DcMDalias_01&format=json2'
        # fetching data
        resp_req = ComDinheiro().requests_api_cd(str_payload)
        json_ = ast.literal_eval(StrHandler().find_between(
            str(resp_req.text.encode('utf8')), "b'", "'").replace(r'\n', ''))
        return JsonFiles().send_json(json_)

    def infos_sectors(self, dt_:datetime) -> json:
        '''
        DOCSTRING: SECTOR INFORMATION REGARDING COMPANIES
        INPUTS: DATA PREGÃO DE INTERESSE
        OUTPUTS: JSON
        '''
        # definindo variável para pool de conexão
        str_payload = 'username={}&password={}&URL=StockScreenerFull.php%3F%26'.format(
            self.user, self.passw) \
            + 'relat%3D%26data_analise%3D{}%2F{}%2F{}%26data_dem%3D31%2F12%2F'.format(
                dt_.strftime('%d'), dt_.strftime('%m'),
                dt_.strftime('%Y')) \
            + '9999%26variaveis%3DTICKER%2BNOME_EMPRESA%2BDATA_REGISTRO%2BSEGMENTO%2BSETOR%' \
            + '2BSUBSETOR%2BSUBSUBSETOR%2BTIPO_BOVESPA({}~{}~{}%2CTODOS%2C%2C)'.format(
                dt_.strftime('%d'), dt_.strftime('%m'),
                dt_.strftime('%Y')
            ) \
            + '%26segmento%3Dtodos%26setor%3Dtodos%26filtro%3D' \
            + '%26demonstracao%3Dconsolidado%2520preferencialmente%26tipo_acao%3D' \
            + 'Todas%26convencao%3DMIXED%26acumular%3D12%26valores_em%3D1%26num_casas' \
            + '%3D2%26salve%3D%26salve_obs%3D%26var_control%3D0%26overwrite%3D0%26' \
            + 'setor_bov%3Dtodos%26subsetor_bov%3Dtodos%26subsubsetor_bov%3Dtodos%26' \
            + 'group_by%3D%26relat_alias_automatico%3DcMDalias_01%26' \
            + 'primeira_coluna_ticker%3D0%26periodos%3D0%26periodicidade%3Danual%26' \
            + 'formato_data%3D1&format=json2'
        # fetching data
        resp_req = ComDinheiro().requests_api_cd(str_payload)
        json_ = ast.literal_eval(StrHandler().find_between(
            str(resp_req.text.encode('utf8')), "b'", "'").replace(r'\n', ''))
        return JsonFiles().send_json(json_)

    @property
    def corporate_events(self) -> json:
        '''
        DOCSTRING: RETORNA INFORMAÇÕES CORPORATIVAS: DIVULGAÇÃO DE RESULTADOS,
        INPUTS: DATA PREGÃO DE INTERESSE
        OUTPUTS: JSON
        '''
        # definindo variável para pool de conexão
        str_payload = 'username={}&password={}&URL=StockScreenerFull.php'.format(
            self.user, self.passw) \
            + '%3F%26relat%3D%26data_analise%3D18%2F06%2F2020%26data_dem%3D' \
            + '31%2F12%2F9999%26variaveis%3DTICKER%2BDATA_ENTREGA_DEM_PRIM%' \
            + '26segmento%3Dtodos%26setor%3Dtodos%26filtro%3D%26demonstracao%3D' \
            + 'consolidado%2520preferencialmente%26tipo_acao%3DTodas%26convencao%3D' \
            + 'MIXED%26acumular%3D12%26valores_em%3D1%26num_casas%3D2%26salve%3D%26' \
            + 'salve_obs%3D%26var_control%3D0%26overwrite%3D0%26setor_bov%3Dtodos%26' \
            + 'subsetor_bov%3Dtodos%26subsubsetor_bov%3Dtodos%26group_by%3D%26' \
            + 'relat_alias_automatico%3DcMDalias_01%26primeira_coluna_ticker%3D0%26' \
            + 'periodos%3D0%26periodicidade%3Danual%26formato_data%3D1&format=json2'
        # fetching data
        resp_req = ComDinheiro().requests_api_cd(str_payload)
        json_ = ast.literal_eval(StrHandler().find_between(
            str(resp_req.text.encode('utf8')), "b'", "'").replace(r'\n', ''))
        return JsonFiles().send_json(json_)

    def trading_infos(self, dt_inf:datetime, dt_sup:datetime) -> json:
        '''
        DOCSTRING: RETURN INFORMATION ABOUT THE AMOUNT OF TRADES, LIQUIDITY ON THE STOCK EXCHANGE,
            MARKET VALUE, TRADING VOLUME, WEIGHT IN THE IBRX100 INDEX, BTC FOR PAPER,
            PURCHASES AND SALES OF FUNDS, FOR STOCKS IN THE IBOV
        INPUTS: DATES (INFERIOR AND SUPERIOR)
        OUTPUTS: JSON
        '''
        # definindo variável para pool de conexão
        str_payload = 'username={}&password={}&URL=StockScreenerFull.php'.format(
            self.user, self.passw) \
            + '%3F%26relat%3D%26data_analise%3D{}%2F{}%2F{}%26data_dem%3D31'.format(
                dt_inf.strftime('%d'), dt_inf.strftime(
                    '%m'), dt_inf.strftime('%Y')
            ) \
            + '%2F12%2F9999%26variaveis%3DTICKER%2B' \
            + 'VOLUME_MEDIO({}~{}~{}%2C{}~{}~{}%2C)'.format(
                dt_inf.strftime('%d'), dt_inf.strftime('%m'),
                dt_inf.strftime('%Y'), dt_sup.strftime('%d'),
                dt_sup.strftime('%m'), dt_sup.strftime('%Y')
            ) \
            + '%2BQUANT_NEGOCIOS({}~{}~{}%2C{}~{}~{}%2C%2Cmedia)%2'.format(
                dt_inf.strftime('%d'), dt_inf.strftime('%m'),
                dt_inf.strftime('%Y'), dt_sup.strftime('%d'),
                dt_sup.strftime('%m'), dt_sup.strftime('%Y')
            ) \
            + 'BMARKET_VALUE%2B' \
            + '%2BLIQUIDEZ_BOLSA({}~{}~{}%2C{}~{}~{}%2C%2C)%2'.format(
                dt_inf.strftime('%d'), dt_inf.strftime('%m'),
                dt_inf.strftime('%Y'), dt_sup.strftime('%d'),
                dt_sup.strftime('%m'), dt_sup.strftime('%Y')
            ) \
            + 'BNEGOCIABILIDADE({}~{}~{}%2C{}~{}~{}%2C%2C%2C2)%2'.format(
                dt_inf.strftime('%d'), dt_inf.strftime('%m'),
                dt_inf.strftime('%Y'), dt_sup.strftime('%d'),
                dt_sup.strftime('%m'), dt_sup.strftime('%Y')
            ) \
            + 'BPESO_INDICE(participacao%2CIBRX%2C{}~{}~{}%2C%2C)%2B'.format(
                dt_sup.strftime('%d'), dt_sup.strftime('%m'),
                dt_sup.strftime('%Y')
            ) \
            + 'BTC_ALUGUEL_ACOES(TV%2C{}~{}~{}%2C{}~{}~{})%2B'.format(
                dt_inf.strftime('%d'),
                dt_inf.strftime('%m'),
                dt_inf.strftime('%Y'),
                dt_sup.strftime('%d'), dt_sup.strftime('%m'),
                dt_sup.strftime('%Y')
            ) \
            + 'COMPRAS_VENDAS_FUNDOS(final_valor%2C{}~{}~{}%2C%2C0)'.format(
                dt_sup.strftime('%d'), dt_sup.strftime('%m'),
                dt_sup.strftime('%Y')
            ) \
            + '%2BPRECO_AJ({}~{}~{}%2C%2C%2CA%2CC)'.format(
                dt_sup.strftime('%d'), dt_sup.strftime('%m'),
                dt_sup.strftime('%Y')
            ) \
            + '%2BDY_12M%26' \
            + 'segmento%3Dtodos%26setor%3Dtodos%26filtro%3D%26demonstracao%3D' \
            + 'consolidado%2520preferencialmente%26tipo_acao%3DTodas%26convencao%3D' \
            + 'MIXED%26acumular%3D12%26valores_em%3D1%26num_casas%3D2%26salve%3D%26' \
            + 'salve_obs%3D%26var_control%3D0%26overwrite%3D0%26setor_bov%3Dtodos%26' \
            + 'subsetor_bov%3Dtodos%26subsubsetor_bov%3Dtodos%26group_by%3D%26' \
            + 'relat_alias_automatico%3DcMDalias_01%26primeira_coluna_ticker%3D' \
            + '0%26periodos%3D0%26periodicidade%3Danual%26formato_data%3D1%26' \
            + 'formato_data%3D1&format=json2'
        # fetching data
        resp_req = ComDinheiro().requests_api_cd(str_payload)
        json_ = ast.literal_eval(StrHandler().find_between(
            str(resp_req.text.encode('utf8')), "b'", "'").replace(r'\n', ''))
        return JsonFiles().send_json(json_)

    def infos_risco(self, dt_inf:datetime, dt_sup:datetime) -> json:
        '''
        DOCSTRING: RETURNS INFORMATION ABOUT 60 MONTHS ANNUALIZED VOLATILITY, VOLATILITY
            ANNUALIZED MONTHLY, YTD ANNUALIZED VOLATILITY, PARAMETRIC VAR, EWMA,
            BENCHMARK VAR PARAMETRIC (IN RELATION TO IBOVESPA), MAXIMUM DRAWDOWN 
            OF STOCKS IN THE IBOV
        INPUTS: DATES (INFERIOR AND SUPERIOR)
        OUTPUTS: JSON
        '''
        # definindo variável para pool de conexão
        str_payload = 'username={}&password={}&URL=StockScreenerFull.php%3F%26'.format(
            self.user, self.passw) \
            + 'relat%3D%26data_analise%3D{}%2F{}%2F{}%26data_dem%3D31%2F12%2F9999'.format(
                dt_sup.strftime('%d'), dt_sup.strftime('%m'),
                dt_sup.strftime('%Y')
            ) \
            + '%26variaveis%3DTICKER%2Bvol_ano_60m%2Bvol_ano_mes_atual%2B' \
            + 'vol_ano_ano_atual%2BVAR_PAR(d%2C{}~{}~{}%2C{}~{}~{}%2C95%2C%2C1)'.format(
                dt_inf.strftime('%d'), dt_inf.strftime('%m'),
                dt_inf.strftime('%Y'), dt_sup.strftime('%d'),
                dt_sup.strftime('%m'), dt_sup.strftime('%Y')
            ) \
            + '%2BVAR_PAR(d%2C{}~{}~{}%2C{}~{}~{}%2C99%2C%2C1)'.format(
                dt_inf.strftime('%d'), dt_inf.strftime('%m'),
                dt_inf.strftime('%Y'), dt_sup.strftime('%d'),
                dt_sup.strftime('%m'), dt_sup.strftime('%Y')
            ) \
            + '%2BEWMA({}~{}~{}%2C{}~{}~{}%2C94%2CB%2C%2C0)%2BBENCHMARK_VAR_PAR('.format(
                dt_inf.strftime('%d'), dt_inf.strftime('%m'),
                dt_inf.strftime('%Y'), dt_sup.strftime('%d'),
                dt_sup.strftime('%m'), dt_sup.strftime('%Y')
            ) \
            + 'd%2C{}~{}~{}%2C{}~{}~{}%2C95%2C%2C1%2CIBOV)%2BMDD'.format(
                dt_inf.strftime('%d'), dt_inf.strftime('%m'),
                dt_inf.strftime('%Y'), dt_sup.strftime('%d'),
                dt_sup.strftime('%m'), dt_sup.strftime('%Y')
            ) \
            + '(d%2C{}~{}~{}%2C{}~{}~{}%2Cmdd)%26segmento%3Dtodos%26setor%3Dtodos'.format(
                dt_inf.strftime('%d'), dt_inf.strftime('%m'),
                dt_inf.strftime('%Y'), dt_sup.strftime('%d'),
                dt_sup.strftime('%m'), dt_sup.strftime('%Y')
            ) \
            + '%26filtro%3D%26demonstracao%3Dconsolidado%2520preferencialmente%26tipo_acao' \
            + '%3DTodas%26convencao%3DMIXED%26acumular%3D12%26valores_em%3D1%26num_casas%3D2' \
            + '%26salve%3D%26salve_obs%3D%26var_control%3D0%26overwrite%3D0%26setor_bov%' \
            + '3Dtodos%26subsetor_bov%3Dtodos%26subsubsetor_bov%3Dtodos%26group_by%3D%26' \
            + 'relat_alias_automatico%3DcMDalias_01%26primeira_coluna_ticker%3D0%26' \
            + 'periodos%3D0%26periodicidade%3Danual%26formato_data%3D1%26' \
            + 'formato_data%3D1&format=json2'
        # fetching data
        resp_req = ComDinheiro().requests_api_cd(str_payload)
        json_ = ast.literal_eval(StrHandler().find_between(
            str(resp_req.text.encode('utf8')), "b'", "'").replace(r'\n', ''))
        return JsonFiles().send_json(json_)

    def trading_volume_securities(self, list_securities:str, str_benchmark:str, dt_inf:datetime, 
                                  dt_sup:datetime) -> json:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # lista de papéis no formato da consulta do json da comdinheiro
        list_securities = '%2B'.join(list_securities).upper()
        # definindo variável para pool de conexão
        str_payload = 'username={}&password={}&URL=ComparaEmpresas001.php'.format(
            self.user, self.passw) \
            + '%3F%26data_d%3D31129999%26data_a%3D{}%2F{}%2F{}%26trailing%3D12%26'.format(
                dt_sup.strftime('%d'), dt_sup.strftime('%m'),
                dt_sup.strftime('%Y')
            ) \
            + 'conv%3DMIXED%26c_c%3Dconsolidado%2520preferencialmente%26moeda%3D' \
            + 'MOEDA_ORIGINAL%26m_m%3D1000000000%26n_c%3D2%26f_v%3D1%26papeis%3D' \
            + '{}'.format(list_securities) \
            + '%26indic%3DTICKER%2BVOLUME_MEDIO%' \
            + '28{}~{}~{}%2C{}~{}~{}%2C%29%2BQUANT_NEGOCIOS%28'.format(
                dt_inf.strftime('%d'), dt_inf.strftime('%m'),
                dt_inf.strftime('%Y'), dt_sup.strftime('%d'),
                dt_sup.strftime('%m'), dt_sup.strftime('%Y')
            ) \
            + '{}~{}~{}%2C{}~{}~{}%2C%2Csoma%29%2B%2BMARKET_VALUE%2BLIQUIDEZ_BOLSA%28'.format(
                dt_inf.strftime('%d'), dt_inf.strftime('%m'),
                dt_inf.strftime('%Y'), dt_sup.strftime('%d'),
                dt_sup.strftime('%m'), dt_sup.strftime('%Y')
            ) \
            + '{}~{}~{}%2C{}~{}~{}%2C%2C%29%2BNEGOCIABILIDADE%28'.format(
                dt_inf.strftime('%d'), dt_inf.strftime('%m'),
                dt_inf.strftime('%Y'), dt_sup.strftime('%d'),
                dt_sup.strftime('%m'), dt_sup.strftime('%Y')
            ) \
            + '{}~{}~{}%2C{}~{}~{}%2C%2C%2C2%29%2BPESO_INDICE%28'.format(
                dt_inf.strftime('%d'), dt_inf.strftime('%m'),
                dt_inf.strftime('%Y'), dt_sup.strftime('%d'),
                dt_sup.strftime('%m'), dt_sup.strftime('%Y')
            ) \
            + 'participacao%2C{}%2C{}~{}~{}%2C%2C%29%2BBTC_ALUGUEL_ACOES'.format(
                str_benchmark, dt_sup.strftime('%d'), dt_sup.strftime(
                    '%m'), dt_sup.strftime('%Y')
            ) \
            + '%28TA%2C{}~{}~{}%2C{}~{}~{}%29%2BCOMPRAS_VENDAS_FUNDOS%28'.format(
                dt_inf.strftime('%d'), dt_inf.strftime('%m'),
                dt_inf.strftime('%Y'), dt_sup.strftime('%d'),
                dt_sup.strftime('%m'), dt_sup.strftime('%Y')
            ) \
            + 'compras_valor%2C{}~{}~{}%2C%2C0%29%2BPRECO_AJ%28'.format(
                dt_sup.strftime('%d'), dt_sup.strftime(
                    '%m'), dt_sup.strftime('%Y')
            ) \
            + '{}~{}~{}%2C%2C%2CA%2CC%29%2BDY_12M%26enviar_email%3D0%26'.format(
                dt_sup.strftime('%d'), dt_sup.strftime(
                    '%m'), dt_sup.strftime('%Y')
            ) \
            + 'enviar_email_log%3D0%26transpor%3D0%26op01%3Dtabela%26' \
            + 'oculta_cabecalho_sup%3D0%26relat_alias_automatico%3DcMDalias_01&format=json2'
        # print('PAYLOAD: {}'.format(str_payload))
        # fetching data
        resp_req = ComDinheiro().requests_api_cd(str_payload)
        # print(resp_req.text)
        json_ = ast.literal_eval(StrHandler().find_between(
            str(resp_req.text.encode('utf8')), "b'", "'").replace(r'\n', ''))
        return JsonFiles().send_json(json_)

    def mdtv_list_securities(
            self, 
            list_securities:List[str], 
            dt_inf:datetime, 
            dt_sup:datetime
        ) -> pd.DataFrame:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # list of securities with the respective format, in order to be used in payload
        list_securities = '%2B'.join(list_securities).upper()
        # payload
        str_payload = 'username={}&password={}&URL=ComparaEmpresas001.php%3F%26data_d%'.format(
            self.user, self.passw) \
            + '3D31129999%26data_a%3D{}%2F{}%2F{}%26trailing%3D12'.format(
                dt_sup.strftime('%d'), dt_sup.strftime('%m'),
                dt_sup.strftime('%Y')
            ) \
            + '%26conv%3DMIXED%26c_c%3Dconsolidado%2520preferencialmente%26moeda%3D' \
            + 'MOEDA_ORIGINAL%26m_m%3D1000000000%26n_c%3D2%26f_v%3D1%26papeis%3D' \
            + '{}'.format(list_securities) \
            + '%26indic%3DVOLUME_MEDIO%28' \
            + '{}~{}~{}%2C{}~{}~{}%'.format(
                dt_inf.strftime('%d'), dt_inf.strftime('%m'),
                dt_inf.strftime('%Y'), dt_sup.strftime('%d'),
                dt_sup.strftime('%m'), dt_sup.strftime('%Y')
            ) \
            + '2C%29%26enviar_email%3D0%26enviar_email_log%3D0%26transpor%3D0%26op01%3Dtabela' \
            + '%26oculta_cabecalho_sup%3D0%26relat_alias_automatico%3DcMDalias_01%26s' \
            + 'cript%3D&format=json2'
        # fetching data
        resp_req = ComDinheiro().requests_api_cd(str_payload)
        json_ = ast.literal_eval(StrHandler().find_between(
            str(str(resp_req.text.encode('utf8')) + "'"), "b'", "''").replace(
                r'\n', '').replace(r' ', ''))
        # return dataframe
        return pd.DataFrame(json_['resposta']['tab-p0']['linha'])

    def risk_infos_list_securities(
            self, 
            list_securities:List[str], 
            str_benchmark:str, 
            dt_inf:datetime, 
            dt_sup: datetime
        ) -> pd.DataFrame:
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # lista de papéis no formato da consulta do json da comdinheiro
        list_securities = '%2B'.join(list_securities).upper()
        # definindo variável para pool de conexão
        str_payload = 'username={}&password={}&URL=ComparaEmpresas001.php%3F%26data_d%'.format(
            self.user, self.passw) \
            + '3D31129999%26data_a%3D{}%2F{}%2F{}%26trailing%3D12'.format(
                dt_sup.strftime('%d'), dt_sup.strftime('%m'),
                dt_sup.strftime('%Y')
            ) \
            + '%26conv%3DMIXED%26c_c%3Dconsolidado%2520preferencialmente%26moeda%3D' \
            + 'MOEDA_ORIGINAL%26m_m%3D1000000000%26n_c%3D2%26f_v%3D1%26papeis%3D' \
            + '{}'.format(list_securities) \
            + '%26indic%3DTICKER%2Bvol_ano_60m%2B%2Bvol_ano_mes_atual%2Bvol_ano_ano_atual%2BVAR_PAR' \
            + '%28d%2C{}~{}~{}%2C{}~{}~{}%2C95%2C%2C1%29%'.format(
                dt_inf.strftime('%d'), dt_inf.strftime('%m'),
                dt_inf.strftime('%Y'), dt_sup.strftime('%d'),
                dt_sup.strftime('%m'), dt_sup.strftime('%Y')
            ) \
            + '2BVAR_PAR%28d%2C{}~{}~{}%2C{}~{}~{}%2C99%2C%2C1%29%2BEWMA'.format(
                dt_inf.strftime('%d'), dt_inf.strftime('%m'),
                dt_inf.strftime('%Y'), dt_sup.strftime('%d'),
                dt_sup.strftime('%m'), dt_sup.strftime('%Y')
            ) \
            + '%28{}~{}~{}%2C{}~{}~{}%2C94%2CB%2C%2C0%29%2BBENCHMARK_VAR_PAR'.format(
                dt_inf.strftime('%d'), dt_inf.strftime('%m'),
                dt_inf.strftime('%Y'), dt_sup.strftime('%d'),
                dt_sup.strftime('%m'), dt_sup.strftime('%Y')
            ) \
            + '%28d%2C{}~{}~{}%2C{}~{}~{}%2C95%2C%2C1%2C'.format(
                dt_inf.strftime('%d'), dt_inf.strftime('%m'),
                dt_inf.strftime('%Y'), dt_sup.strftime('%d'),
                dt_sup.strftime('%m'), dt_sup.strftime('%Y')
            ) \
            + '{}%29%2BMDD%28d%2C'.format(str_benchmark) \
            + '{}~{}~{}%2C26~01~2021%2Cmdd%29%26enviar_email%3D0'.format(
                dt_inf.strftime('%d'), dt_inf.strftime('%m'),
                dt_inf.strftime('%Y')
            ) \
            + '%26enviar_email_log%3D0%26transpor%3D0%26op01%3Dtabela' \
            '%26oculta_cabecalho_sup%3D0%26relat_alias_automatico%3DcMDalias_01&format=json2'
        # fetching data
        resp_req = ComDinheiro().requests_api_cd(str_payload)
        json_ = ast.literal_eval(StrHandler().find_between(
            str(resp_req.text.encode('utf8')), "b'", "'").replace(r'\n', ''))
        return JsonFiles().send_json(json_)

    def stocks_beta(self, list_tickers:List[str], dt_inf:datetime, dt_sup:datetime) -> json:
        '''
        DOCSTRING: BETA OF PROVIDED STOCKS
        INPUTS:  LIST OF TICKER AND INFERIOR AND SUPERIOR DATES
        OUTPUTS: JSON
        '''
        # lista de papéis no formato da consulta do json da comdinheiro
        list_securities = '%2B'.join(list_tickers)
        # applying date format
        dt_inf = DatesBR().datetime_to_string(dt_inf, '%d%m%Y')
        dt_sup = DatesBR().datetime_to_string(dt_sup, '%d%m%Y')
        # payload
        str_payload = \
            'username={}&password={}&URL=HistoricoIndicadoresFundamentalistas001.php'.format(
                self.user, self.passw
            ) \
            + '%3F%26data_ini%3D{}%26data_fim%3D{}'.format(dt_inf, dt_sup) \
            + '%26trailing%3D12%26conv%3DMIXED%26moeda%3D' \
            + 'BRL%26c_c%3Dconsolidado%26m_m%3D1000000%26n_c%3D2%26f_v%3D1%26papel%3D{}'.format(
                list_securities) \
            + '%26indic%3Dret_01d%2Bret_cdi_01d%2Bbeta_06m%2BLC%26periodicidade%3Ddu%26graf_' \
            + 'tab%3Dtabela%26desloc_data_analise%3D1%26flag_transpor%3D0%26c_d%3Dd%26enviar_email' \
            + '%3D0%26enviar_email_log%3D0%26' \
            + 'cabecalho_excel%3Dmodo1%26relat_alias_automatico%3DcMDalias_01&format=json2'
        # fetching data
        json_ = ComDinheiro().requests_api_cd(
            str_payload).read()
        json_ = ast.literal_eval(str(json_.decode('utf8').replace(
            r'\n', '').replace(r' ', '')))
        return JsonFiles().send_json(json_)

    def tradability_index(self, list_tickers:List[str], dt_inf:datetime, dt_sup:datetime) -> json:
        '''
        DOCSTRING: 
        INPUTS:
        OUTPUTS:
        '''
        # lista de papéis no formato da consulta do json da comdinheiro
        list_securities = '%2B'.join(list_tickers)
        # applying date format
        dt_inf = DatesBR().datetime_to_string(dt_inf, '%d%m%Y')
        dt_sup = DatesBR().datetime_to_string(dt_sup, '%d%m%Y')
        # definindo variável para pool de conexão
        str_payload = 'username={}&password={}&URL=ComparaFundos001.php%3F%26datas'.format(
            self.user, self.passw) \
            + '%3D{}%2F{}%2F{}%26cnpjs'.format(dt_sup.strftime('%d'), dt_sup.strftime('%m'), 
                dt_sup.strftime('%Y')) \
            + '%3D{}'.format(list_securities) \
            + '%26indicadores%3DNEGOCIABILIDADE%28{}~{}~{}%2C{}~{}~{}'.format(
                dt_inf.strftime('%d'), dt_inf.strftime('%m'), dt_inf.strftime('%Y'), 
                dt_sup.strftime('%d'), dt_sup.strftime('%m'), dt_sup.strftime('%Y')) \
            + '%2C%2C%2C2%29%26num_casas%3D2%26pc%3Dnome_fundo%26flag_transpor%3D0%26enviar_email' \
            + '%3D0%26mostrar_da%3D0%26op01%3Dtabela%26oculta_cabecalho_sup%3D0%26r' \
            + 'elat_alias_automatico%3DcMDalias_01&format=json2'
        # fetching data
        json_ = ComDinheiro().requests_api_cd(str_payload).read()#.text.encode('utf8')
        json_ = ast.literal_eval(str(json_.decode('utf8').replace(
            r'\n', '').replace(r' ', '')))
        # returning data
        return JsonFiles().send_json(json_)
    
    def open_ended_funds_quotes(self, list_funds:List[str], dt_inf:datetime, dt_sup:datetime) -> json:
        '''
        DOCSTRING: CLOSING PRICE OF FUNDS' SHARE
        INPUTS:  FUNDS CNPJ (list), MATURITY CODE,
            DATE INFERIOR (DDMMAAAsA, AS A STRING, OR DATETIME FORMAT), DATE SUPERIOR (SAME FORMAT
            AS DATE INFERIOR) AND FORMAT EXTRACTION (JSON AS DEFAULT)
        OUTPUTS: JSON
        '''
        # list of securities with the format requested to be queried
        list_securities = '%2B'.join(list_funds)
        # applying date format
        if DatesBR().check_date_datetime_format(dt_inf) == True:
            dt_inf = DatesBR().datetime_to_string(dt_inf, '%d%m%Y')
        if DatesBR().check_date_datetime_format(dt_sup) == True:
            dt_sup = DatesBR().datetime_to_string(dt_sup, '%d%m%Y')
        # payload
        str_payload = 'username={}&password={}&URL=HistoricoIndicadoresFundos001.php'.format(
            self.user, self.passw) \
            + '%3F%26cnpjs%3D{}'.format(list_securities) \
            + '%26data_ini%3D{}%26data_fim%3D{}'.format(dt_inf, dt_sup) \
            + '%26indicadores%3Dvalor_cota%26op01%3Dtabela_h%26num_casas%3D2%26enviar_email%3D0' \
            + '%26periodicidade%3Ddiaria%26cabecalho_excel%3Dmodo1%26transpor%3D0%26asc_desc%3Ddesc' \
            + '%26tipo_grafico%3Dlinha%26relat_alias_automatico%3DcMDalias_01&format=json2'
        # fetching data
        json_ = ComDinheiro().requests_api_cd(
            str_payload).read()
        json_ = ast.literal_eval(json_.decode('utf8').replace(r'\n', '').replace(
            r' ', ''))
        return JsonFiles().send_json(json_)

    def open_ended_funds_risk_infos(self, list_funds:List[str], dt_sup:datetime) -> json:
        '''
        DOCSTRING: RISK INFOS REGARDING HISTORICAL VOLATILITY AND REDEMPTION
        INPUTS:  FUNDS CNPJ (list), DATE SUPERIOR
        OUTPUTS: JSON
        '''
        # list of securities with the format requested to be queried
        list_securities = '%2B'.join(list_funds)
        # applying date format
        dt_sup = DatesBR().datetime_to_string(dt_sup, '%d%m%Y')
        # payload
        str_payload = 'username={}&password={}&URL=HistoricoIndicadoresFundos001.php'.format(
            self.user, self.passw) \
            + '%3F%26cnpjs%3D{}'.format(list_securities) \
            + '%26data_ini%3D{}%26data_fim%3D{}'.format(dt_sup, dt_sup) \
            + '%26indicadores%3Dprazo_liq_resg%2B' \
            + 'prazo_disp_rec_resgatado%2Bvol_ano_01m%2Bvol_ano_12m%2B' \
            + 'vol_ano_36m%2Bvol_ano_60m%2Bvol_ano_ano_atual%2Bvol_ano_mes_atual%2Bresgate_min%2Btaxa_saida%26op01%3Dtabela_h%26' \
            + 'num_casas%3D2%26enviar_email%3D0%26periodicidade%3Ddiaria%26cabecalho_excel%3Dmodo1%26transpor%3D0%26asc_desc%3D' \
            + 'desc%26tipo_grafico%3Dlinha%26relat_alias_automatico%3DcMDalias_01&format=json2'
        # fetching data
        json_ = ComDinheiro().requests_api_cd(
            str_payload).read()
        json_ = ast.literal_eval(json_.decode('Latin-1').replace(
            r'\n', '').replace(r' ', ''))
        return JsonFiles().send_json(json_)

    def open_ended_funds_sharpe_dd(
            self, 
            list_eins:List[str], 
            dt_inf:datetime, 
            dt_sup:datetime
        ) -> json:
        '''
        DOCSTRING: CÁLCULO DE SHARPE E DROWDOWN POR PERÍODO
        INPUTS:
        OUTPUTS:
        '''
        # list of securities with the format requested to be queried
        list_eins = '%2B'.join(list_eins)
        # payload
        str_payload = 'username={}&password={}&URL=ComparaFundos001.php%3F%26datas'.format(
            self.user, self.passw) \
            + '%3D{}%2F{}%2F{}%26cnpjs'.format(dt_sup.strftime('%d'), dt_sup.strftime('%m'), 
                dt_sup.strftime('%Y')) \
            + '%3D{}'.format(list_eins) \
            + '%26indicadores%3Dnome_fundo%2Bcnpj_fundo%2Bret_12m_aa%2Bpatrimonio~1e6%2B' \
            + 'cotistas%2Bcaptacao~1e6%2Bresgate~1e6%2' \
            + 'Bvol_ano_24m%2Bsharpe_12m%2BMDD%28d%2C{}~{}~{}'.format(
                dt_inf.strftime('%d'), dt_inf.strftime('%m'), dt_inf.strftime('%Y')) \
            + '%2C{}~{}~{}%2Cmdd%29%'.format(
                dt_sup.strftime('%d'), dt_sup.strftime('%m'), dt_sup.strftime('%Y')) \
            + '2BMDD%28d%2C{}~{}~{}%'.format(
                dt_inf.strftime('%d'), dt_inf.strftime('%m'), dt_inf.strftime('%Y')) \
            + '2C{}~{}~{}%2'.format(
                dt_sup.strftime('%d'), dt_sup.strftime('%m'), dt_sup.strftime('%Y')) \
            + 'Ctempo_rec%29%2Bobjetivo%2Bclasse%26num_casas%3D2%26pc%3Dnome_fundo%26f' \
            + 'lag_transpor%3D0%26enviar_email%3D0%26mostrar_da%3D0%26op01%3Dtabela%26' \
            + 'oculta_cabecalho_sup%3D0%26relat_alias_automatico%3DcMDalias_01&format=json2'
        # fetching data
        json_ = ComDinheiro().requests_api_cd(str_payload).text.encode('utf8')#.read()
        json_ = ast.literal_eval(str(json_.decode('utf8').replace(
            r'\n', '').replace(r' ', '')))
        # returning data
        return JsonFiles().send_json(json_)
    
    def open_ended_funds_infos(
            self, 
            list_eins:List[str], 
            dt_inf:datetime, 
            dt_sup:datetime, 
            str_fmt_dt_input:str='YYYY-MM-DD', 
            int_wd_bef:int=2,
            col_classe_anbima:str='CLASSE_ANBIMA_DO_FUNDO', 
            col_nome_fundo:str='NOME_FUNDO', 
            col_like_max_drawdown:str='MDD*', 
            col_mdd_final:str='MAX_DRAWDOWN', 
            col_cnpj_sem_mascara:str='CNPJ_SEM_MASCARA', 
            col_cnpj:str='CNPJ_DO_FUNDO'
        ) -> pd.DataFrame:
        '''
        DOCSTRING: FUND'S NAME, CNPJ, CORPORATE NAME, SHARPE WITH A RANGE OF TEMPORAL WINDOWS, 
            MAXIMUM DRAWDOWN (GIVEN A PERIOD OF TIME), ANBIMA'S CLASS, CODE
        INPUTS: LIST OF EINS, DATE INFERIOR AND SUPERIOR
        OUTPUTS: DATAFRAME
        '''
        # setting variables
        list_ser = list()
        # changing dates format
        dt_inf = DatesBR().str_date_to_datetime(dt_inf, str_fmt_dt_input)
        dt_sup = DatesBR().str_date_to_datetime(dt_sup, str_fmt_dt_input)
        # inferior reference date within comdinheiro
        dt_ref_cf = DatesBR().sub_working_days(dt_sup, int_wd_bef)
        # eins with the format requested to be queried
        list_eins = '%2B'.join(list_eins)
        # payload
        str_payload = 'username={}&password={}&URL=ComparaFundos001.php%3F%26'.format(
            self.user, self.passw) \
            + 'datas%3D{}%2F{}%2F{}%26'.format(dt_ref_cf.strftime('%d'), dt_ref_cf.strftime('%m'), 
                dt_ref_cf.strftime('%Y')) \
            + 'cnpjs%3D{}%26'.format(list_eins) \
            + 'indicadores%3Dcnpj_fundo%2Bnome_fundo%2Bsharpe_12m%2Bsharpe_24m%2Bsharpe_' \
            + '36m%2Bsharpe_36m%2Bsharpe_48m%2Bsharpe_60m' \
            + '%2BMDD%28d%2C{}~{}~{}%2C{}~{}~{}%2Cmdd%29'.format(
                dt_inf.strftime('%d'), dt_inf.strftime('%m'), dt_inf.strftime('%Y'), 
                dt_sup.strftime('%d'), dt_sup.strftime('%m'), dt_sup.strftime('%Y')
            ) \
            + '%2Bclasse_anbima%2Bcodigo_anbima%26num_casas%3D2%26pc%3Dnome_fundo%26flag_transpor' \
            + '%3D0%26enviar_email%3D0%26mostrar_da%3D0%26op01%3Dtabela%26oculta_cabecalho_sup%3D0' \
            + '%26relat_alias_automatico%3DcMDalias_01&format=json3'
        # fetching data
        json_message = ComDinheiro().requests_api_cd(str_payload).json()
        # cols of exportation
        list_cols = [StrHandler().remove_diacritics(
            StrHandler().latin_characters(str(x))).strip().replace(' ', '_').replace(
            dt_ref_cf.strftime('%d/%m/%Y'), '').upper() for x in list(json_message[
                'tables']['tab0']['lin0'].values())]
        # serialized list of dictionaries
        for index, dict_ in json_message['tables']['tab0'].items():
            if index != 'lin0':
                list_ser.append({col_nome: dict_['col{}'.format(i)] for i, 
                                   col_nome in enumerate(list_cols)})
        # serialized list to pandas dataframe
        df_infos_fundos_cd = pd.DataFrame(list_ser)
        # replacing null values
        for col_ in list_cols:
            df_infos_fundos_cd[col_] = [str(x).replace('', '0') if len(str(x)) <= 1 
                                        else str(x) for x in df_infos_fundos_cd[col_].tolist()]
            df_infos_fundos_cd[col_] = [str(x).replace(',', '.') for x in df_infos_fundos_cd[
                col_].tolist()]
        # changing columns data types
        df_infos_fundos_cd = df_infos_fundos_cd.astype({
            list_cols[0]: str,
            list_cols[1]: str,
            list_cols[2]: str,
            list_cols[3]: float,
            list_cols[4]: float,
            list_cols[5]: float,
            list_cols[6]: float,
            list_cols[7]: float,
            list_cols[8]: float,
            list_cols[9]: float,
            list_cols[10]: str,
            list_cols[11]: int
        })
        # renaming columns of interest
        col_mdd = [x for x in list_cols if StrHandler().match_string_like(
            x, col_like_max_drawdown) == True][0]
        df_infos_fundos_cd.rename(columns={
            col_mdd: col_mdd_final
        }, inplace=True)
        # removing diacritics
        for col_ in [col_nome_fundo, col_classe_anbima]:
            df_infos_fundos_cd[col_] = [StrHandler().remove_diacritics(
                StrHandler().latin_characters(str(x))).upper() 
                for x in df_infos_fundos_cd[col_].tolist()]
        # ein without mask
        df_infos_fundos_cd[col_cnpj_sem_mascara] = DocumentsNumbersBR(
            df_infos_fundos_cd[col_cnpj].tolist()).unmask_number
        # returning pandas dataframe
        return df_infos_fundos_cd



### DADOS DE NEGOCIAÇÃO DE DÍVIDA PÚBLICA BRASILEIRA ###

import datetime
import pandas as pd
from io import BytesIO
from requests import request
import sys
sys.path.append(r'C:\Users\Guilherme\OneDrive\Dev\Python\Packages')
from stpstone.handling_data.str import StrHandler
from stpstone.central._global_slots import YAML_BR_TRS
from stpstone.cals.handling_dates import DatesBR


class ConsultaDadosTesouroDireto:

    def pu_taxa_abertura_contratos(self, contrato, data_mtm, data_vencimento):
        '''
        REFERENCES: https://sisweb.tesouro.gov.br/apex/f?p=2031:2:0:
        DOCSTRING: PU E TAXA NA ABERTURA DO DIA
        INPUTS: CONTRATO, DATA DE REFERÊNCIA PARA MARCAÇÃO À MERCADO E DATA DE VENCIMENTO
        OUTPUTS: DICIONÁRIO (DIA, TAXA DE COMPRA NA ABERTURA, TAXA DE VENDA NA ABERTURA, 
            PU COMPRA NA ABERTURA, PU VENDA NA ABERTURA, PU VENDA NA ABERTURA, PU BASE ABERTURA)
        '''
        # ajustando tipos da variáveis
        if isinstance(data_mtm, datetime.date) == True:
            data_mtm = DatesBR().datetime_to_string(data_mtm)
        if isinstance(data_vencimento, datetime.date) == True:
            data_vencimento = DatesBR().datetime_to_string(data_vencimento)
        # datas de referência para a solicitação
        data_vencimento = DatesBR().str_date_to_datetime(data_vencimento)
        ano_vencimento = DatesBR().str_date_to_datetime(data_mtm).strftime('%Y')
        data_vencimento = data_vencimento.strftime('%d%m%y')
        # requisição de tabela do tesouro direto com o contrato de interesse para o ano de
        #   mtm de referência
        url = 'https://sisweb.tesouro.gov.br/apex/cosis/sistd/obtem_arquivo/{}'.format(
            YAML_BR_TRS['sisweb_td_host_port'][contrato][ano_vencimento])
        response = request('GET', url).content
        # importando para dataframe o contrato com vencimento de interesse
        with BytesIO(response) as fh:
            df_contrato_vencimento = pd.io.excel.read_excel(
                fh, sheet_name='{} {}'.format(contrato, data_vencimento))
        new_header = df_contrato_vencimento.iloc[0]
        df_contrato_vencimento = df_contrato_vencimento[1:]
        df_contrato_vencimento.columns = new_header
        # coletando taxa e pu para a data de referência
        for index, row in df_contrato_vencimento.iterrows():
            if df_contrato_vencimento.loc[index, 'Dia'] == data_mtm:
                return {
                    'dia': df_contrato_vencimento.loc[index, 'Dia'],
                    'taxa_compra_abertura': df_contrato_vencimento.loc[index, 'Taxa Compra Manhã'],
                    'taxa_venda_abertura': df_contrato_vencimento.loc[index, 'Taxa Venda Manhã'],
                    'pu_compra_abertura': df_contrato_vencimento.loc[index, 'PU Compra Manhã'],
                    'pu_venda_abertura': df_contrato_vencimento.loc[index, 'PU Venda Manhã'],
                    'pu_base_abertura': df_contrato_vencimento.loc[index, 'PU Base Manhã'],
                }
        else:
            raise Exception('Data de referência para marcação a mercado não encontrada para o '
                            + 'vencimento')


# print(ConsultaDadosTesouroDireto().pu_taxa_abertura_contratos(
#     'NTN-B', '27/11/2020', '15/08/2024'))
# output
# {'dia': '27/11/2020', 'taxa_compra_abertura': 0.0196, 'taxa_venda_abertura': 0.0208,
#     'pu_compra_abertura': 3917.08, 'pu_venda_abertura': 3901.7, 'pu_base_abertura': 3898.38}
# print(ConsultaDadosTesouroDireto().pu_taxa_abertura_contratos(
#     'LTN', DatesBR().sub_working_days(DatesBR().curr_date(), 1), '01/01/2021'))
# output
# {'dia': '27/11/2020', 'taxa_compra_abertura': 0.0196, 'taxa_venda_abertura': 0.0208,
#     'pu_compra_abertura': 998.22, 'pu_venda_abertura': 998.12, 'pu_base_abertura': 998.04}
# print(ConsultaDadosTesouroDireto().pu_taxa_abertura_contratos(
#     'LTN', DatesBR().sub_working_days(DatesBR().curr_date(), 1), '01/01/2021'))
# # output
# {'dia': '27/11/2020', 'taxa_compra_abertura': 0.0196, 'taxa_venda_abertura': 0.0208,
#     'pu_compra_abertura': 998.22, 'pu_venda_abertura': 998.12, 'pu_base_abertura': 998.04}
# print(ConsultaDadosTesouroDireto().pu_taxa_abertura_contratos(
#     'NTN-C', DatesBR().sub_working_days(DatesBR().curr_date(), 1), '01/04/2021'))
# # output
# {'dia': '27/11/2020', 'taxa_compra_abertura': -0.032799999999999996, 'taxa_venda_abertura': -
#     0.0316, 'pu_compra_abertura': 5240.43, 'pu_venda_abertura': 5238.26, 'pu_base_abertura': 5222.04}
# print(ConsultaDadosTesouroDireto().pu_taxa_abertura_contratos(
#     'NTN-B', DatesBR().sub_working_days(DatesBR().curr_date(), 1), '15/08/2040'))
# # output
# {'dia': '27/11/2020', 'taxa_compra_abertura': 0.0404, 'taxa_venda_abertura': 0.0416,
#     'pu_compra_abertura': 4321.25, 'pu_venda_abertura': 4259.12, 'pu_base_abertura': 4255.17}
# print(ConsultaDadosTesouroDireto().pu_taxa_abertura_contratos(
#     'NTN-B', DatesBR().sub_working_days(DatesBR().curr_date(), 1), '15/08/2050'))
# # output
# {'dia': '27/11/2020', 'taxa_compra_abertura': 0.042300000000000004, 'taxa_venda_abertura': 0.0435,
#     'pu_compra_abertura': 4435.09, 'pu_venda_abertura': 4355.19, 'pu_base_abertura': 4351.11}
# print(ConsultaDadosTesouroDireto().pu_taxa_abertura_contratos(
#     'NTN-B', DatesBR().sub_working_days(DatesBR().curr_date(), 500), '15/08/2050'))
# {'dia': '03/12/2018', 'taxa_compra_abertura': 0.05, 'taxa_venda_abertura': 0.0512, 'pu_compra_abertura': 3713.98, 'pu_venda_abertura': 3649.07, 'pu_base_abertura': 3648.44}
# print(ConsultaDadosTesouroDireto().pu_taxa_abertura_contratos(
#     'NTN-B', '17/02/2021', '15/08/2024'))
# # output
# {'dia': '17/02/2021', 'taxa_compra_abertura': 0.0191, 'taxa_venda_abertura': 0.0203, 'pu_compra_abertura': 3923.48, 'pu_venda_abertura': 3908.67, 'pu_base_abertura': 3907.44}



from urllib.parse import urlencode


def buildUrlWithParams(url: str, params={}) -> str:
    encoded = urlencode(cleanDict(params))
    return url if (len(encoded) == 0) else (url + '?' + encoded)


def cleanDict(x: dict = {}) -> dict:
    result = {}
    for key in x:
        if x[key] is not None:
            # Check If List
            if isinstance(x[key], (list,)):
                result[key] = ','.join(x[key])
            # Check If Boolean
            elif isinstance(x[key], (bool)):
                if x[key] is True:
                    result[key] = 'true'
                else:
                    result[key] = 'false'
            # Everything Else (Strings/Numbers)
            else:
                result[key] = x[key]
    return result


def mergeDict(x: dict, y: dict) -> dict:
    z = cleanDict(x)
    z.update(cleanDict(y))
    return z



### API TO GRANT ACCESS TO B3 MARGIN SIMULATOR ###

from requests import request


class MarginSimulatorB3:

    def __init__(self, client_id, client_secret, token=None,
                 hostname_api_margin_simulator_b3='https://api.b3.com.br/cors-app'):
        self.client_id = client_id
        self.client_secret = client_secret
        self.token = token
        self.hostname_api_margin_simulator_b3 = hostname_api_margin_simulator_b3

    @property
    def access_token(self, method='GET', bl_verify=False,
                     bl_str_dict_params=False, app='/api/oauth/token'):
        '''
        DOCSTRING: ACCESS TOKEN TO GRANT ACCESS TO MARGIN SIMULATOR B3
        INPUTS: METHOD (DEFAULT), KEY HEADER (DEFAULT), URL AUTHENTIFICATION
            HEADER (DEFAULT)
        OUTPUTS: STRING
        '''
        # passing variables
        i = 0
        # requesting authorization authheader
        dict_headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
        }
        # parameters
        dict_params = {
            'grant_type': 'client_credentials',
            'client_id': str(self.client_id),
            'client_secret': str(self.client_secret),
        }
        #   coverting to string, if its user will
        if bl_str_dict_params == True:
            dict_params = '&'.join('{}={}'.format(k, v)
                                   for k, v in dict_params.items())
        # looping while status code not a 2xx response
        resp_req = request(method=method, url=self.hostname_api_margin_simulator_b3 + app,
                           headers=dict_headers, params=dict_params, verify=bl_verify)
        # raises exception when not a 2xx response
        resp_req.raise_for_status()
        # getting authheader
        return resp_req.json()



### TERM STRUCTURE OF INTEREST RATE CURVE ###

import matplotlib.pyplot as plt
from stpstone.handling_data.folders import DirFilesManagement


class TSIR:

    def nper_ytm_curve(self, dict_nper_rates, chart_title, complete_path_save_fig,
                       flag_show_plot=False):
        '''
        REFERENCES: https://matplotlib.org/tutorials/introductory/pyplot.html
        DOCSTRING: YTM CURVE
        INPUTS: DICT NPER RATES, CHART TITLE, COMPLETE PATH TO SAVE THE FIGURE AND BOOLEAN 
            TO WHETER SHOW OR NOT THE CHART (FALSE AS STANDARD)
        OUTPUTS:
        '''
        plt.figure()
        plt.plot(dict_nper_rates.keys(), dict_nper_rates.values())
        plt.ylabel('Yield to Maturity')
        plt.xlabel('Working Days to Maturity')
        plt.title(chart_title)
        plt.savefig(complete_path_save_fig)
        if flag_show_plot == True:
            plt.show()
        return DirFilesManagement().object_exists(complete_path_save_fig)



### HANDLING TXT FILES ###

from re import sub
from stpstone.handling_data.folders import DirFilesManagement


class HandlingTXTFiles:

    def read_file(self, complete_path, method='r'):
        '''
        DOCSTRING: READ A TXT FILE
        INPUTS: COMPLETE PATH AND METHOD (R, FROM READ, AS DEFAULT)
        OUTPUTS: CONTENT
        '''
        with open(complete_path, method) as f:
            return f.read()

    def generator_file(self, complete_path, method='r', regex='[^A-Za-z0-9-.*-:-<-=-\s]+',
                       non_matching_regex_characaters='|'):
        '''
        DOCSTRING: 
        INPUTS:
        OUTPUTS:
        '''
        with open(complete_path, method, encoding="ascii", errors='replace') as f:
            return [sub(regex, non_matching_regex_characaters, line) for line in
                    f.read().splitlines()]

    def read_first_line(self, complete_path, method='r'):
        '''
        DOCSTRING: READ THE FIRST LINE OF A TXT FILE
        INPUTS: COMPLETE PATH AND METHOD (R, FROM READ, AS DEFAULT)
        OUTPUTS: CONTENT
        '''
        with open(complete_path, method) as f:
            return f.readline().rstrip()

    def remove_first_n_lines(self, complete_path, n=1):
        '''
        DOCSTRING: REMOVE THE FIRST N LINES OF A TXT FILE
        INPUTS: COMPLETE PATH AND N FROM NUMBER OF LINES (1 AS DEFAULT) 
        OUTPUTS: ORIGINAL CONTENT WITHOUT THE FIRST N LINES
        '''
        with open(complete_path, 'r') as file_in:
            data = file_in.read().splitlines(True)
        with open(complete_path, 'w') as file_out:
            file_out.writelines(data[n:])

    def write_file(self, file_complete_path, data_content, method='w'):
        '''
        DOCSTRING: WRITE A TXT FILE
        INPUTS: FILE COMPLETE PATH, DATA CONTENT AND METHOD (DEFAULT AS W THAT STANDS FOR WRITE)
        OUTPUTS: STATUS ACCOMPLISHMENT
        '''
        with open(file_complete_path, method) as file_output:
            file_output.write(data_content)
        return DirFilesManagement().object_exists(file_complete_path)



### USA WORKING DAYS ###

from datetime import datetime, timedelta
from workalendar.usa import UnitedStates


class WorkCalendar:

    def __init__(self) -> None:
        '''
        DOCSTRING: INITIALIZES THE WORK CALENDAR
        INPUTS: NONE
        OUTPUTS: NONE
        '''
        self.calendar = UnitedStates()

    def add_working_days(self, dt_bgn: datetime, num_days: int) -> datetime:
        '''
        DOCSTRING: ADDS A SPECIFIED NUMBER OF WORKING DAYS TO A GIVEN START DATE
        INPUTS: DT_START (START DATE), NUM_DAYS (NUMBER OF WORKING DAYS TO ADD)
        OUTPUTS: DT_END (END DATE)
        '''
        current_date = dt_bgn
        working_days_added = 0
        while working_days_added < num_days:
            current_date += timedelta(days=1)
            if current_date.weekday() < 5 and not self.calendar.is_holiday(current_date):
                working_days_added += 1
        return current_date

    def is_holiday(self, date: datetime) -> bool:
        ''' 
        DOCSTRING: CHECKS IF A GIVEN DATE IS A HOLIDAY IN THE USA
        INPUTS: DATE (AS DATETIME)
        OUTPUTS: BOOLEAN
        '''
        return self.calendar.is_holiday(date)

    def is_weekend(self, date: datetime) -> bool:
        ''' 
        DOCSTRING: CHECKS IF A GIVEN DATE IS A WEEKEND (SATURDAY OR SUNDAY) 
        INPUTS: DATE (AS DATETIME)
        OUTPUTS: BOOLEAN
        '''
        return date.weekday() >= 5
    
    def diff_working_days(self, dt_bgn: datetime, dt_end: datetime) -> int:
        ''' 
        DOCSTRING: CALCULATES THE NUMBER OF WORKING DAYS BETWEEN TWO DATES
        INPUTS: START DATE (AS DATETIME), END DATE (AS DATETIME)
        OUTPUTS: INTEGER
        '''
        current_date = dt_bgn
        working_days = 0
        while current_date <= dt_end:
            if current_date.weekday() < 5 and not self.calendar.is_holiday(current_date):
                working_days += 1
            current_date += timedelta(days=1)
        return working_days


if __name__ == '__main__':

    cls_usabzdays = WorkCalendar()
    # Current date
    dt_bgn = datetime(2025, 1, 1)
    # Add 10 working days to the start date
    new_date = cls_usabzdays.add_working_days(dt_bgn, 10)
    print(f"Date after adding 10 working days: {new_date}")
    # Check if a specific date is a holiday
    check_date = datetime(2025, 1, 20)  # Martin Luther King Jr. Day (USA)
    print(f"Is {check_date} a holiday? {cls_usabzdays.is_holiday(check_date)}")
    # Check if a specific date is a weekend
    weekdt_end = datetime(2025, 1, 11)
    print(f"Is {weekdt_end} a weekend? {cls_usabzdays.is_weekend(weekdt_end)}")



import typing
from .client import Client


class Forms:
    """Typeform Forms API client"""

    def __init__(self, client: Client):
        """Constructor for Typeform Forms class"""
        self.__client = client
        self.__messages = FormMessages(client)

    @property
    def messages(self):
        return self.__messages

    def create(self, data: dict = {}) -> dict:
        """Creates a form"""
        return self.__client.request('post', '/forms', data=data)

    def delete(self, uid: str) -> str:
        """
        Deletes the form with the given form_id and all of the form's responses.
        Return a `str` based on success of deletion, `OK` on success, otherwise an error message.
        """
        return self.__client.request('delete', '/forms/%s' % uid)

    def get(self, uid: str) -> dict:
        """Retrieves a form by the given form_id. Includes any theme and images attached to the form as references."""
        return self.__client.request('get', '/forms/%s' % uid)

    def list(self, page: int = None, pageSize: int = None, search: str = None,
             workspaceId: str = None) -> dict:
        """
        Retrieves a list of JSON descriptions for all forms in your Typeform account (public and private).
        Forms are listed in reverse-chronological order based on the last date they were modified.
        """
        return self.__client.request('get', '/forms', params={
            'page': page,
            'page_size': pageSize,
            'search': search,
            'workspace_id': workspaceId
        })

    def update(self, uid: str, data: dict = {}, patch: bool = False) -> typing.Union[str, dict]:
        """
        Updates an existing form.
        Defaults to `put`.
        `put` will return the modified form as a `dict` object.
        `patch` will return a `str` based on success of change, `OK` on success, otherwise an error message.
        """
        methodType = 'put' if patch is False else 'patch'
        return self.__client.request(methodType, '/forms/%s' % uid, data=data)


class FormMessages:
    def __init__(self, client: Client):
        """Constructor for TypeForm FormMessages class"""
        self.__client = client

    def get(self, uid: str) -> dict:
        """
        Retrieves the customizable messages for a form (specified by form_id) using the form's specified language.
        You can format messages with bold (*bold*) and italic (_italic_) text. HTML tags are forbidden.
        """
        return self.__client.request('get', '/forms/%s/messages' % uid)

    def update(self, uid: str, data={}) -> str:
        """
        Specifies new values for the customizable messages in a form (specified by form_id).
        You can format messages with bold (*bold*) and italic (_italic_) text. HTML tags are forbidden.
        Return a `str` based on success of change, `OK` on success, otherwise an error message.
        """
        return self.__client.request('put', '/forms/%s/messages' % uid, data=data)



# HANDLE CALCULUS ISSUES WITH PYTHON

import random
import sympy as sym
import pandas as np
from scipy.integrate import trapz, cumtrapz, quad
from linear_algebra import distance, add, scalar_multiply


class Calculus:

    def variables(self, str_symbols):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return sym.symbols(str_symbols.split(' '))

    def differentiation(self, f, variable_, nth_derivative=1, *args_symbols):
        '''
        DOCSTRING: PERFORM NTH DIFFERENTIATION OF A FUNCTION
        INPUTS: FUNCTION, VARIABLE OF DIFFERENTIATION, NTH DERIVATIVE, *ARGS SYMBOLS
        OUTPUTS: POLYNOM
        '''
        return sym.diff(f, variable_, nth_derivative)

    def integration(self, f, variable_, lower_bound=None, upper_bound=None):
        '''
        DOCSTRING: PERFORM NTH DIFFERENTIATION OF A FUNCTION
        INPUTS: FUNCTION, VARIABLE, LOWER BOUNDARY, UPPER BOUNDARY
        OUTPUTS: POLYNOM
        '''
        if all([x == None for x in [lower_bound, upper_bound]]):
            return sym.integrate(f, variable_)
        else:
            return sym.integrate(f, (variable_, lower_bound, upper_bound))

    def trapz_integration(self, f, variable_):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return trapz(f, variable_)

    def cumtrapz_integration(self, f, variable_):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return cumtrapz(f, variable_)

    def simplify(self, f):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return sym.simplify(f)

    def gradient_step(self, array_indep, gradient, step_size):
        '''
        DOCSTRING: MOVES STEP_SIZE IN THE GRADIENT DIRECTION FROM V
        INPUTS: VECTOR, GRADIENT, STEP_SIZE
        OUTPUTS: VECTOR
        '''
        array_indep = np.array(array_indep)
        assert len(array_indep) == len(gradient)
        step = scalar_multiply(step_size, gradient)
        return add(array_indep, step)

    def sum_of_squares_gradient(self, array_indep):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return [2 * v_i for v_i in array_indep]

    def least_gradient_vector(self, step_size=-0.01, iter=1000, epsilon=0.001):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # pick a random starting point
        array_indep = [random.uniform(-10, 10) for i in range(3)]
        # loop through
        for epoch in range(iter):
            # compute the gradient at array_indep
            grad = self.sum_of_squares_gradient(array_indep)
            # take a negative gradient step
            array_indep = self.gradient_step(array_indep, grad, step_size)
            print(epoch, array_indep)
        # array_indep should be close to 0
        assert distance(array_indep, [0, 0, 0]) < epsilon
        # returning array with least gradient vector
        return array_indep


# # differentiation
# x, y, z = Calculus().variables('x y z')
# f = x * y + x ** 2 + sym.sin(2 * y)
# df_dx = Calculus().differentiation(f, x, 1, x, y, z)
# print(df_dx)
# # output: 2*x + y

# # integration
# x, y, z = Calculus().variables('x y z')
# f = x * y + x ** 2 + sym.sin(2 * y)
# F = Calculus().integration(f, x, 0, 2)
# print(F)
# # output: 2*y + 2*sin(2*y) + 8/3

# # integration
# x, y, z = Calculus().variables('x y z')
# f = x * y + x ** 2 + sym.sin(2 * y)
# F = Calculus().integration(f, x)
# print(F)
# # output: x**3/3 + x**2*y/2 + x*sin(2*y)



### COLLECT WORLD GOVERNMENT BONDS DATA ###

import pandas as pd
from stpstone.settings._global_slots import YAML_GLB_RT
from stpstone.handling_data.html import HtmlHndler
from stpstone.handling_data.dicts import HandlingDicts
from stpstone.handling_data.str import StrHandler
from stpstone.loggs.db_logs import DBLogs
from stpstone.cals.handling_dates import DatesBR
from stpstone.handling_data.numbers import NumHandler


class GlobalRates:

    def generic_req(self, url, bl_verify=False):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        # setting variables
        list_outlooks = list()
        # request html
        bs_html = HtmlHndler().html_bs_parser(
            url, 
            bl_verify
        )
        # table
        bs_table = bs_html.find('table')
        # headers
        list_th = [
            YAML_GLB_RT['general']['col_rate']
            if len(
                th.get_text()
                    .replace('\n', '')
                    .replace('/', '_')
                    .strip()
            ) == 0
            else
            th.get_text()
                .replace('\n', '')
                .replace('/', '_')
                .strip()
                .upper()
            for th in bs_table.find_all('th')
        ]
        # data
        list_td = [
            #   turning into percentage values that are numbers
            float(td.get_text()
                .replace('\n', '')
                .replace('%', '')
                .strip()) / 100.0
            if NumHandler().is_numeric(
                td.get_text()
                    .replace('\n', '')
                    .replace('%', '')
                    .strip()
            )
            #   return value otherwise
            else
            td.get_text()
                .replace('\n', '')
                .replace('%', '')
                .strip()
            for td in bs_table.find_all('td')
            if len(
                td.get_text()
                    .replace('\n', '')
                    .replace('%', '')
                    .replace('-', '0')
                    .strip()
            ) > 0
        ]
        # checking wheter the table has direction info
        try:
            #   outlooks
            for div in bs_table.find_all('div', class_='table-normal text-end'):
                try:
                    str_outlook = StrHandler().replace_all(
                        div.find('i')['class'][1], 
                        YAML_GLB_RT['central_banks_rates']['rate_outlook_desc']
                    )
                    list_outlooks.append(str_outlook)
                except (AttributeError, TypeError):
                    continue
            #   input direction in the data dictionary
            int_col_dir = list_th.index(YAML_GLB_RT['central_banks_rates']['col_direction'])
            int_rng_upper = len(list_td) // (len(list_th) - 1)
            for i in range(int_rng_upper):
                list_td.insert(3 + (int_col_dir + 3) * i, list_outlooks[i])
        except ValueError:
            pass
        # creating list of dictionaries
        list_ser = HandlingDicts().pair_headers_with_data(
            list_th, 
            list_td
        )
        # turning into dataframe
        df_ = pd.DataFrame(list_ser)
        # adding logging
        df_ = DBLogs().audit_log(
            df_, 
            DatesBR().utc_from_dt(DatesBR().curr_date), 
            DatesBR().utc_log_ts
        )
        # return dataframe
        return df_

    @property
    def libor(self):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return self.generic_req(
            YAML_GLB_RT['libor_rates']['url'], 
            YAML_GLB_RT['libor_rates']['bl_verify']
        )

    @property
    def euribor(self):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return self.generic_req(
            YAML_GLB_RT['euribor_rates']['url'], 
            YAML_GLB_RT['euribor_rates']['bl_verify']
        )
    
    @property
    def central_banks(self):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return self.generic_req(
            YAML_GLB_RT['central_banks_rates']['url'], 
            YAML_GLB_RT['central_banks_rates']['bl_verify']
        )
    
    @property
    def ester(self):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return self.generic_req(
            YAML_GLB_RT['ester']['url'], 
            YAML_GLB_RT['ester']['bl_verify']
        )
    
    @property
    def sonia(self):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return self.generic_req(
            YAML_GLB_RT['sonia']['url'], 
            YAML_GLB_RT['sonia']['bl_verify']
        )
    
    @property
    def sofr(self):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return self.generic_req(
            YAML_GLB_RT['sofr']['url'], 
            YAML_GLB_RT['sofr']['bl_verify']
        )
    
    @property
    def usa_cpi(self):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return self.generic_req(
            YAML_GLB_RT['usa_cpi']['url'], 
            YAML_GLB_RT['usa_cpi']['bl_verify']
        )
    
    @property
    def uk_cpi(self):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return self.generic_req(
            YAML_GLB_RT['uk_cpi']['url'], 
            YAML_GLB_RT['uk_cpi']['bl_verify']
        )
    
    @property
    def can_cpi(self):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return self.generic_req(
            YAML_GLB_RT['can_cpi']['url'], 
            YAML_GLB_RT['can_cpi']['bl_verify']
        )
    
    @property
    def eur_cpi(self):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return self.generic_req(
            YAML_GLB_RT['eur_cpi']['url'], 
            YAML_GLB_RT['eur_cpi']['bl_verify']
        )


### HANDLING NUMERICAL ISSUES ###

import math
import operator
import functools
from fractions import Fraction
from math import gcd
from stpstone.handling_data.str import StrHandler


class NumHandler:
    '''
    REFERENCES: http://www.hoadley.net/options/develtoolsvolcalc.htm,
        https://introcs.cs.princeton.edu/python/21function/blackscholes.py.html,
        https://aaronschlegel.me/implied-volatility-functions-python.html#:~:text=Implied%20volatility%20σim,can%27t%20be%20directly%20observed.
    '''

    def multiples(self, m, closest_ceiling_num):
        '''
        DOCSTRING: LIST OF NUMERICAL MULTIPLES FROM A GIVEN NUMBER
        INPUTS: MULTIPLE AND THE CLOSEST CEILING NUMBER (ROUNDED UP)
        OUTPUTS: LIST
        '''
        # appending multiples
        list_numerical_mulptiples = list()
        count = int(closest_ceiling_num / m) + 2
        for i in range(0, count * m, m):
            list_numerical_mulptiples.append(i)
        # replacing last value
        if list_numerical_mulptiples[-1] > closest_ceiling_num:
            list_numerical_mulptiples[-1] = closest_ceiling_num
        # output
        return list_numerical_mulptiples

    def nearest_multiple(self, number, multiple):
        '''
        DOCSTRING: RETURN THE NEAREST MULTIPLE OF A GIVEN NUMBER
        INPUTS: NUMBER AND MULTIPLE
        OUTPUTS: INTEGER
        '''
        return multiple * int(number / multiple)

    def round_up(self, float_number_to_round, float_base, float_ceiling):
        '''
        DOCSTRING: ROUND UP A DIVISION WITH A CEILING
        INPUTS: FLOAT NUMERATOR, FLOAT DENOMINATOR, AND FLOAT CEILING
        OUTPUTS: FLOAT
        '''
        # correcting variables to float type
        float_number_to_round, float_base, float_ceiling = (float(x) for x in
                                                            [float_number_to_round, float_base, float_ceiling])
        # defining next multiple with a ceiling
        if float(float_base + self.truncate(float_number_to_round / float_base, 0)
                 * float_base) < float_ceiling:
            return float(float_base + self.truncate(float_number_to_round / float_base, 0)
                         * float_base)
        else:
            return float_ceiling

    def decimal_to_fraction(self, decimal_number):
        '''
        DOCSTRING: FRACTION FROM A DECIMAL
        INPUTS: DECIMAL NUMBER
        OUTPUTS: FRACTION OF A NUMBER
        '''
        return Fraction(decimal_number)

    def greatest_common_divisor(self, int1, int2):
        '''
        DOCSTRING: GREATEST COMMON DIVISOR BETWEEN TWO INTEGERS
        INPUTS: INTEGER 1 AND INTEGER 2
        OUTPUTS: GREATEST COMMON DIVISOR
        '''
        return gcd(int1, int2)

    def truncate(self, number, digits):
        '''
        DOCSTRING: TRUNCATE A NUMBER IN NTH-DECIMAL
        INPUTS: NUMBER AND DIGITS
        OUTPUTS: FLOAT
        '''
        stepper = 10.0 ** digits
        return math.trunc(stepper * number) / stepper

    def sumproduct(self, *lists):
        '''
        REFERENCES: https://stackoverflow.com/questions/3849251/sum-of-products-for-multiple-lists-in-python
        DOCSTRING: SUMPRODUCT, OR POSITIONAL MULTIPLIACTION OF N LISTS
        INPUTS: *LISTS
        OUTPUTS: FLOAT
        '''
        return sum(functools.reduce(operator.mul, data) for data in zip(*lists))

    def number_sign(self, number, base_number=1):
        '''
        DOCSTRING: SIGN OF A GIVEN NUMBER
        INPUTS: NUMBER AND BASE (1 AS DEFAULT)
        OUTPUTS: EITHER 1 OR -1
        '''
        return math.copysign(base_number, number)

    def multiply_n_elements(self, *args):
        '''
        DOCSTRING: MULTIPLY A GIVEN SET OF ARGUMENTS
        INPUTS: ELEMENTS TO BE MULTIPLIED
        OUTPUTS: A GIVEN SET OF DATA MULTIPLIED, IN THE SAME FORMAT AS THE INPUT
        '''
        product = 1
        for a in args:
            product *= a
        return product

    def sum_n_elements(self, *args):
        '''
        DOCSTRING: SUM A GIVEN SET OF ARGUMENTS
        INPUTS: ELEMENTS TO BE ADDED
        OUTPUTS: A GIVEN SET OF DATA ADDED, IN THE SAME FORMAT AS THE INPUT
        '''
        sum_ = 0
        for a in args:
            sum_ += a
        return sum_

    def factorial(self, n):
        '''
        DOCSTRING: FACTORIAL MATHEMATICAL FUNCTION
        INPUTS: INTEGER N
        OUTPUTS: INTEGER
        '''
        return functools.reduce(operator.mul, range(1, n + 1))

    def convert_thousands_decimals_separator(self, number_float, precision_decimals=2,
                                             thousands_inputs='.', thousands_outputs=',',
                                             decimals_inputs=',', decimals_outputs='.'):
        '''
        DOCSTRING: CONVERTING THOUSANDS AND DECIMALS SEPARATORS
        INPUTS: NUMBER FLOAT AND NUMBER FORMAT
        OUTPUTS: NUMBER FLOAT WITH DESIRED FORMAT
        '''
        # defining number format output string
        number_format_output = ':{}{}{}f'.format(thousands_outputs, decimals_outputs,
                                                 precision_decimals)
        number_format_output = '{' + number_format_output + '}'
        # converting to float if number is string
        if type(number_float) == str:
            number_int = ''.join(str(StrHandler().get_string_until_substr(
                number_float, decimals_inputs)).split(thousands_inputs))
            number_decimals = StrHandler().get_string_after_substr(
                number_float, decimals_inputs)
            number_float = float(
                number_int + decimals_outputs + number_decimals)
        # converting to interested format
        return {
            'number_float': number_float,
            'str_number_formated': number_format_output.format(number_float)
        }

    def range_floats(self, float_epsilon, float_inf, float_sup, float_pace):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS:
        '''
        return [float(x) / float_epsilon for x in range(int(float_inf * float_epsilon), 
                                                        int(float_sup * float_epsilon), 
                                                        int(float_pace * float_epsilon))]

    def clamp(self, n, minn, maxn):
        '''
        DOCSTRING: CONSTRICT NUMBER WITHIN RANGE - WINSORIZE
        INPUTS:
        OUTPUTS:
        '''
        return max(min(maxn, n), minn)

    def is_numeric(self, str_):
        '''
        DOCSTRING:
        INPUTS:
        OUTPUTS
        '''
        try:
            float(str_)
            return True
        except ValueError:
            return False


### HANDLING TGZ FILES ###

import tarfile
import os
from stpstone.handling_data.folders import DirFilesManagement


class HandlingTGZFiles:

    def fetch_tgz_files(self, dir_exporting_path, tgz_exporting_name, url_source=None,
                        complete_source_path=None):
        '''
        REFERENCES: "HANDS-ON MACHINE LEARNING WITH SCIKIT-LEARN, KERAS, AND TENSORFLOW, 
            2ND EDITION, BY AURÉLIEN GÉRON (O’REILLY). COPYRIGHT 2019 KIWISOFT S.A.S., 
            978-1-492-03264-9.”
        DOCSTRING: EXTRACT FILES FROM TGZ COMPRESSION
        INPUTS: DIRECTORY EXPORTING PATH, TGZ FILE EXPORTING NAME, URL SOURCE (WHEN APPLICABLE), 
            COMPLETE SOURCE PATH (IF THE FILE IS ALREADY IN THE DISK)
        OUTPUTS: DICTIONARY WITH BLAME DOWNLOAD TGZ FILE, DIR PATH AND EXTRACTED FILES NAMES
        '''
        # checking wheter the directory to export the file is available or not
        if DirFilesManagement().object_exists(dir_exporting_path):
            DirFilesManagement().mk_new_directory(dir_exporting_path)
        # exporting complete path
        tgz_path = os.path.join(dir_exporting_path, tgz_exporting_name)
        # print(tgz_path)
        # requesting data when it is provided a valid url, whereas a complete source path have
        #   to be declared
        if all([x == None for x in [url_source, complete_source_path]]):
            raise Exception('Url source or complete path of source ought be passed in order to '
                            + 'data from the file.')
        elif url_source != None:
            blame_download_tgz_file = DirFilesManagement().download_web_file(
                url_source, tgz_path)
        else:
            blame_download_tgz_file = 'n/a'
        # extracting data from tgz file
        tgz_file = tarfile.open(tgz_path)
        tgz_members_names = tgz_file.getnames()
        tgz_file.extractall(path=dir_exporting_path)
        tgz_file.close()
        # return whether or not the file exists
        return {
            'blame_download_tgz_file': blame_download_tgz_file,
            'dir_path': dir_exporting_path,
            'extracted_files_names': tgz_members_names
        }



### INOA SYSTEMS ###

import json
import backoff
import pandas as pd
from requests import request, exceptions
from typing import List
from datetime import datetime
from stpstone.settings._global_slots import YAML_INOA
from stpstone.cals.handling_dates import DatesBR


class AlphaTools:

    def __init__(self, str_user:str, str_passw:str, str_host:str, str_instance:str, 
        dt_start:datetime, dt_end:datetime, str_fmt_date_output:str='YYYY-MM-DD', 
        bl_debug_mode:bool=False):
        '''
        DOCSTRING: CONNECTION TO INOA SERVICE ALPHATOOLS
        INPUTS: USER, PASSWORD, HOST, INSTANCE, LIST OF EINS, START AND END DATES
        OUTPUTS: -
        '''
        self.str_user = str_user
        self.str_passw = str_passw
        self.str_host = str_host
        self.str_instance = str_instance
        self.dt_start = dt_start
        self.dt_end = dt_end
        self.str_fmt_date_output = str_fmt_date_output
        self.bl_debug_mode = bl_debug_mode if bl_debug_mode is not None else True

    @backoff.on_exception(
        backoff.constant,
        exceptions.RequestException,
        interval=10,
        max_tries=20,
    )
    def generic_req(self, str_method:str, str_app:str, dict_params:dict) -> json:
        '''
        DOCSTRING: GENERIC REQUEST TO ALPHA TOOLS API
        INPUTS: METHOD, APP, PARAMETERS
        OUTPUTS: JSON
        '''
        if self.bl_debug_mode == True:
            print(
                f'METHOD: {str_method}',
                f'HOST: {self.str_host}',
                f'APP: {str_app}',
                f'DICT PARAMS: {dict_params}',
                f'USER: {self.str_user}',
                f'PASSW:{self.str_passw}'
            )
        resp_req = request(str_method, url=self.str_host+str_app, json=dict_params, 
            auth=(self.str_user, self.str_passw))
        resp_req.raise_for_status()
        return resp_req.json()

    @property
    def funds(self) -> pd.DataFrame:
        '''
        DOCSTRING: FUNDS INFOS, WITHIN A GIVEN SET
        INPUTS: - 
        OUTPUTS: DATAFRAME
        '''
        # setting initial variables
        dict_params = {
            YAML_INOA['alpha_tools']['funds']['key_values']: [
                YAML_INOA['alpha_tools']['funds']['col_id'], 
                YAML_INOA['alpha_tools']['funds']['col_name'], 
                YAML_INOA['alpha_tools']['funds']['col_legal_id']
            ],
            YAML_INOA['alpha_tools']['funds']['key_is_active']: None
        }
        # requesting infos
        json_req = self.generic_req(
            YAML_INOA['alpha_tools']['funds']['method'],
            YAML_INOA['alpha_tools']['funds']['app'], 
            dict_params
        )
        # retrieving to dataframe in memory
        df_funds = pd.DataFrame.from_dict(json_req, orient='index')
        if self.bl_debug_mode == True:
            print(f'DF_FUNDS: \n{df_funds}')
        # changing columns types
        df_funds = df_funds.astype({
            YAML_INOA['alpha_tools']['funds']['col_id']: int,
            YAML_INOA['alpha_tools']['funds']['col_name']: str,
            YAML_INOA['alpha_tools']['funds']['col_legal_id']: str
        })
        # including columns of interest
        df_funds[YAML_INOA['alpha_tools']['funds']['col_origin']] = self.str_instance
        # changing columns case
        df_funds.columns = [x.upper() for x in df_funds.columns]
        # returning dataframe of funds
        return df_funds

    def quotes(self, list_ids:List[int]) -> pd.DataFrame:
        '''
        DOCSTRING: FUNDS' QUOTES
        INPUTS: -
        OUTPUTS: DATAFRAME
        '''
        # setting initial variables
        dict_params = {
            YAML_INOA['alpha_tools']['quotes']['key_funds_ids']: list_ids,
            YAML_INOA['alpha_tools']['quotes']['key_start_dt']: self.dt_start.strftime(
                '%Y-%m-%d'),
            YAML_INOA['alpha_tools']['quotes']['key_end_dt']: self.dt_end.strftime(
                '%Y-%m-%d'),
        }
        # requesting infos
        json_req = self.generic_req(
            YAML_INOA['alpha_tools']['quotes']['method'],
            YAML_INOA['alpha_tools']['quotes']['app'], 
            dict_params
        )
        # print(f'JSON_QUOTES_AT: {json_req}')
        # retrieving to dataframe in memory
        df_quotes = pd.DataFrame(json_req[
            YAML_INOA['alpha_tools']['quotes']['key_items']])
        # changing columns types
        if self.bl_debug_mode == True:
            print(f'DF_QUOTES: \n{df_quotes}')
        df_quotes = df_quotes.astype({
            YAML_INOA['alpha_tools']['quotes']['col_fund_id']: int,
            YAML_INOA['alpha_tools']['quotes']['col_date']: str,
            YAML_INOA['alpha_tools']['quotes']['col_status_display']: str
        })
        df_quotes[YAML_INOA['alpha_tools']['quotes']['col_date']] = [
            DatesBR().str_date_to_datetime(d, self.str_fmt_date_output) 
            for d in df_quotes[YAML_INOA['alpha_tools']['quotes']['col_date']]
        ]
        # changing columns case
        df_quotes.columns = [x.upper() for x in df_quotes.columns]
        # returning dataframe
        return df_quotes


from setuptools import setup, find_packages

with open('README.md') as fp:
    long_description = fp.read()

setup(
    name='typeform',
    version='1.1.0',
    description='Python Client wrapper for Typeform API',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=[
        'type',
        'form',
        'typeform',
        'api',
    ],
    author='Typeform',
    author_email='michael.solati@typeform.com',
    url='https://github.com/MichaelSolati/typeform-python-sdk',
    packages=find_packages(),
    install_requires=['requests'],
    test_suite='typeform.test.suite.test_suite',
    license='MIT',
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python',
    ]
)



### INFOS FROM CEI ENVIRONMENT OF INVESTOR'S TRADES IN BRAZILLIAN B3 MARKET ###

import asyncio
from bolsa import B3AsyncBackend
from pprint import pprint
from stpstone.cals.handling_dates import DatesBR


class CEIB3:

    def __init__(self, username, password, captcha_service=None):
        '''
        INPUTS: USERNAME (CPF WITHOUT . AND -), PASSWORD AND CAPTCHA SERVICE (NONE AS DEFAULT, 
            SINCE ITS NOT A MUST YET)
        '''
        self.username = username
        self.password = password
        self.captcha_service = captcha_service

    async def extract_data(self):
        '''
        REFERENCES: https://github.com/gicornachini/bolsa/blob/master/examples/simple.py
        DOCSTRING: EXTRACT DATA FROM CEI B3
        INPUTS: -
        OUTPUTS: ASSETS EXTRACTED IN A TUPLE
        '''
        # setting local variables
        dict_assets_extracted = dict()
        b3_httpclient = B3AsyncBackend(
            username=self.username,
            password=self.password,
            captcha_service=None  # captcha_service is not required yet
        )
        # defining brokers with available accounts in b3
        brokers = await b3_httpclient.get_brokers_with_accounts()
        dict_assets_extracted['brokers'] = brokers
        # deffining custody in each broker
        dict_assets_extracted['custody'] = (
            await b3_httpclient.get_brokers_account_portfolio_assets_extract(
                brokers=brokers
            )
        )
        # closing connection
        await b3_httpclient.session_close()
        await b3_httpclient.connection_close()
        # returning available brokers and custody in each one
        return dict_assets_extracted

    def assets_under_custody(self):
        '''
        DOCSTRING: ASSETS UNDER CUSTODY IN CEI B3
        INPUTS: -
        OUTPUTS: DICTIONARY
        '''
        # setting local variables
        dict_export = dict()
        dict_assets = dict()
        i_asset = 0
        # async threading to run extract data function inside CEI B3 class
        if __name__ == '__main__':
            # openning event loop
            loop = asyncio.get_event_loop()
            # running extract data function
            dict_assets_extracted = loop.run_until_complete(
                CEIB3(self.username, self.password, self.captcha_service).extract_data())
            # placing custody in each available broker
            dict_assets_extracted['reference_date'] = dict_assets_extracted[
                'brokers'][0].parse_extra_data.end_date
            for i in range(len(dict_assets_extracted['custody'])):
                # defining broker number in list
                if dict_assets_extracted['brokers'][i].name != '1982 - MODAL DTVM LTDA':
                    i_broker = i
                else:
                    i_broker = i + 1
                # broker info
                dict_export[dict_assets_extracted['brokers'][i_broker].name] = {
                    'broker_number': dict_assets_extracted['brokers'][i_broker].value,
                    'account': dict_assets_extracted['brokers'][i_broker].accounts[0].id,
                }
                # checking wheter the custody list is empty, and continue otherwise
                if str(dict_assets_extracted['custody'][i]) != '[]':
                    # extracting assets infos of the current broker
                    for asset in dict_assets_extracted['custody'][i]:
                        # since its possible to have several operations in the same asset, place
                        #   the ticker with a counter
                        if asset.raw_negotiation_code + '_' + str(i_asset) not in \
                                dict_assets.keys():
                            # reseting local variable
                            i_asset = 0
                            # creating key name for the main dictionary
                            key = asset.raw_negotiation_code + \
                                '_' + str(i_asset)
                        else:
                            i_asset += 1
                            key = asset.raw_negotiation_code + \
                                '_' + str(i_asset)
                        # asset infos
                        dict_assets[key] = {
                            'asset_name': asset.asset_specification,
                            'market_type': asset.market_type,
                            'operation_date': DatesBR().datetime_to_string(asset.operation_date),
                            'action': asset.action,
                            'quantity': asset.unit_amount,
                            'price': float(asset.unit_price),
                            'notional': float(asset.total_price),
                            'quotation_factor': asset.quotation_factor
                        }
                    # placing in export dictionary infos of the assets under management in
                    #   the current broker, available for the user account
                    dict_export[dict_assets_extracted['brokers'][i + 1].name]['custody'] = \
                        dict_assets
                    # exit loop if its in Modal DTVM due to its custody return being denied
                    if dict_assets_extracted['brokers'][i].name == '1982 - MODAL DTVM LTDA':
                        break
                else:
                    dict_export['custody'] = dict_assets_extracted['custody'][i]
            # returning required data
            return dict_export

    def earnings(self):
        '''
        REFERENCES: https://github.com/wagnervielmond/ProventosB3/blob/main/getProventosCEI.py
        DOCSTRING: EARNINGS RECEIVED IN BRAZILLIAN STOCK MARKET
        INPUTS: -
        OUTPUTS: DICTIONARY
        '''
        # setting local variables
        dict_export = dict()
        dict_assets = dict()
        i_asset = 0
        # async threading to run extract data function inside CEI B3 class
        if __name__ == '__main__':
            # openning event loop
            loop = asyncio.get_event_loop()
            # running extract data function
            dict_assets_extracted = loop.run_until_complete(
                CEIB3(self.username, self.password, self.captcha_service).extract_data())
            # placing custody in each available broker
            dict_assets_extracted['reference_date'] = dict_assets_extracted[
                'brokers'][0].parse_extra_data.end_date
            for i in range(len(dict_assets_extracted['custody'])):
                # defining broker number in list
                if dict_assets_extracted['brokers'][i].name != '1982 - MODAL DTVM LTDA':
                    i_broker = i
                else:
                    i_broker = i + 1
                # broker info
                dict_export[dict_assets_extracted['brokers'][i_broker].name] = {
                    'broker_number': dict_assets_extracted['brokers'][i_broker].value,
                    'account': dict_assets_extracted['brokers'][i_broker].accounts[0].id,
                }
                # checking wheter the custody list is empty, and continue otherwise
                if str(dict_assets_extracted['custody'][i]) != '[]':
                    # extracting assets infos of the current broker
                    for asset in dict_assets_extracted['custody'][i]:
                        # since its possible to have several operations in the same asset, place
                        #   the ticker with a counter
                        if asset.raw_negotiation_code + '_' + str(i_asset) not in \
                                dict_assets.keys():
                            # reseting local variable
                            i_asset = 0
                            # creating key name for the main dictionary
                            key = asset.raw_negotiation_code + \
                                '_' + str(i_asset)
                        else:
                            i_asset += 1
                            key = asset.raw_negotiation_code + \
                                '_' + str(i_asset)
                        # asset infos
                        dict_assets[key] = {
                            'asset_name': asset.asset_specification,
                            'raw_negotiation_code': asset.raw_negotiation_code,
                            'operation_date': DatesBR().datetime_to_string(asset.operation_date),
                            'event_type': asset.event_type,
                            'unit_amount': asset.unit_amount,
                            'quotation_factor': int(asset.quotation_factor),
                            'bruto_price': float(asset.bruto_price),
                            'liquido_price': asset.liquido_price
                        }
                    # placing in export dictionary infos of the assets under management in
                    #   the current broker, available for the user account
                    dict_export[dict_assets_extracted['brokers'][i + 1].name]['custody'] = \
                        dict_assets
                    # exit loop if its in Modal DTVM due to its custody return being denied
                    if dict_assets_extracted['brokers'][i].name == '1982 - MODAL DTVM LTDA':
                        break
                else:
                    dict_export['custody'] = dict_assets_extracted['custody'][i]
            # returning required data
            return dict_export
